"""``bsk_rl.utils.orbital``:Utilities for computing orbital events."""

import logging
from typing import Callable, Iterable, Optional, Tuple, Union

import numpy as np
from Basilisk import __path__
from Basilisk.simulation import eclipse, spacecraft
from Basilisk.utilities import (
    SimulationBaseClass,
    macros,
    orbitalMotion,
    simIncludeGravBody,
)
from Basilisk.utilities.orbitalMotion import ClassicElements, elem2rv, rv2elem
from scipy.interpolate import interp1d

bskPath = __path__[0]

logger = logging.getLogger(__name__)


def random_unit_vector() -> np.ndarray:
    """Generate a random unit vector.

    Returns:
        Random unit vector
    """
    vec = np.random.randn(3)
    return vec / np.linalg.norm(vec)


def random_orbit(
    i: Optional[float] = None,
    a: Optional[float] = 6371 + 500,
    e: float = 0,
    Omega: Optional[float] = None,
    omega: Optional[float] = None,
    f: Optional[float] = None,
    alt: float = None,
    r_body: float = 6371,
) -> ClassicElements:
    """Create a set of orbit elements.

    Parameters are fixed if specified and randomized if ``None``. Defaults to a random
    circular orbit at 500 km altitude.

    Args:
        i: [deg] Inclination, randomized in ``[-pi, pi]``.
        a: [km] Semi-major axis.
        e: Eccentricity.
        Omega: [deg] LAN, randomized in ``[0, 2pi]``.
        omega: [deg] Argument of periapsis, randomized in ``[0, 2pi]``.
        f: [deg] True anomaly, randomized in ``[0, 2pi]``.
        alt: [km] (deprecated) Altitude above r_body.
        r_body: [km] (deprecated) Body radius.

    Returns:
        ClassicElements: orbital elements
    """
    if alt is not None:
        logger.warning(
            "Ignoring a, e, and omega and using alt and r_body to generate a circular orbit."
            "random_circular_orbit is preferred for this use case."
        )
        return random_circular_orbit(
            i=i,
            alt=alt,
            r_body=r_body,
            Omega=Omega,
            f=f,
        )

    oe = ClassicElements()
    oe.a = a * 1e3
    oe.e = e
    oe.i = np.radians(i) if i is not None else np.random.uniform(-np.pi, np.pi)
    oe.Omega = (
        np.radians(Omega) if Omega is not None else np.random.uniform(0, 2 * np.pi)
    )
    oe.omega = (
        np.radians(omega) if omega is not None else np.random.uniform(0, 2 * np.pi)
    )
    oe.f = np.radians(f) if f is not None else np.random.uniform(0, 2 * np.pi)
    return oe


def random_circular_orbit(
    i: Optional[float] = None,
    alt: float = 500,
    r_body: float = 6371,
    Omega: Optional[float] = None,
    f: Optional[float] = None,
):
    """Create a set of orbit elements for a circular orbit.

    Parameters are fixed if specified and randomized if ``None``.

    Args:
        i: [deg] Inclination, randomized in ``[-pi, pi]``.
        alt: [km] Altitude above r_body.
        r_body: [km] Body radius, defaults to Earth's radius.
        Omega: [deg] LAN, randomized in ``[0, 2pi]``.
        f: [deg] True anomaly, randomized in ``[0, 2pi]``.

    """
    omega = 0
    e = 0
    a = r_body + alt

    return random_orbit(i=i, a=a, e=e, Omega=Omega, omega=omega, f=f)


def random_epoch(start: int = 2000, end: int = 2022):
    """Generate a random epoch in a year range.

    Date will always be in the first 28 days of the month.

    Args:
        start: Initial year.
        end: Final year.

    Returns:
        Epoch in ``YYYY MMM DD HH:MM:SS.SSS (UTC)`` format
    """
    year = np.random.randint(start, end)
    month = np.random.choice(
        [
            "JAN",
            "FEB",
            "MAR",
            "APR",
            "MAY",
            "JUN",
            "JUL",
            "AUG",
            "SEP",
            "OCT",
            "NOV",
            "DEC",
        ]
    )
    day = np.random.randint(1, 28)  # Assume 28 days for simplicity
    hours = np.random.randint(0, 23)
    minutes = np.random.randint(0, 59)
    seconds = np.random.randint(0, 59)
    milliseconds = np.random.randint(0, 999)

    # Combine the parts to form the datetime string
    epoch = (
        f"{year} {month} {day:02d} "
        + f"{hours:02d}:{minutes:02d}:{seconds:02d}.{milliseconds:03d} (UTC)"
    )

    return epoch


def elevation(r_sat: np.ndarray, r_target: np.ndarray) -> np.ndarray:
    """Find the elevation angle from a target to a satellite.

    Args:
        r_sat: Satellite position(s).
        r_target: Target position.

    Returns:
        Elevation angle(s)
    """
    if r_sat.ndim == 2:
        return np.pi / 2 - np.arccos(
            np.sum(r_target * (r_sat - r_target), 1)
            / (np.linalg.norm(r_target) * np.linalg.norm(r_sat - r_target, axis=1))
        )
    else:
        return np.pi / 2 - np.arccos(
            np.sum(r_target * (r_sat - r_target))
            / (np.linalg.norm(r_target) * np.linalg.norm(r_sat - r_target))
        )


def walker_delta(
    n_spacecraft: int,
    n_planes: int,
    rel_phasing: float,
    altitude: float,
    inc: float,
    clustersize: int = 1,
    clusterspacing: float = 0,
) -> list[orbitalMotion.ClassicElements]:
    """Compute the initial orbit conditions of a Walker-delta constellation.

    Args:
        n_spacecraft: Number of spacecraft.
        n_planes: Number of orbital planes.
        rel_phasing: [deg] Relative phasing between planes.
        altitude: [m] Altitude above Earth's surface.
        inc: [deg] Inclination.
        clustersize: Number of spacecraft in each cluster.
        clusterspacing: [deg] Spacing between spacecraft in a cluster.

    Returns:
        list: List of orbital elements
    """
    oe_all = []

    # Loop through s/c
    for idx in range(0, n_spacecraft):
        # Instantiate orbital elements
        oe = orbitalMotion.ClassicElements()

        # Define altitude, eccentricity, and inclination
        oe.a = 6371 * 1000.0 + altitude
        oe.e = 0.0
        oe.i = np.radians(inc)

        # Define the plane number
        spacecraft_per_plane = float(n_spacecraft / n_planes)
        plane_num = float(int(idx / spacecraft_per_plane))

        # Compute longitude of ascending node
        oe.Omega = np.radians(plane_num * 360.0 / n_planes % 360)

        # Set argument of periapsis using relative phasing
        dPhi_rel = plane_num * rel_phasing * 360.0 / n_planes
        oe.omega = np.radians(dPhi_rel)

        # Define true anomoly using in-plane phasing
        dPhi_inplane = 360.0 / (spacecraft_per_plane / clustersize)
        oe.f = np.radians(
            (int((idx % spacecraft_per_plane) / clustersize) * dPhi_inplane)
            + clusterspacing * (idx % clustersize)
        )

        # Append to oe_all
        oe_all.append(oe)

    return oe_all


def walker_delta_args(
    n_planes: int,
    rel_phasing: float = 0.0,
    altitude: float = 500,
    inc: float = 45.0,
    randomize_true_anomaly=True,
    randomize_lan=True,
    **walker_delta_kwargs,
):
    """Generate a function to generate a randomly phased Walker-delta constellation.

    The output of this function should be used to set the ``sat_arg_randomizer`` of the
    environment.

    Args:
        n_planes: Number of constellation planes.
        rel_phasing: [deg] Relative phasing between planes. Defaults to 0.0.
        altitude: [km] Orbit altitude above Earth's surface.
        inc: [deg] Orbit inclination
        randomize_true_anomaly: Add a random true anomaly phase shift that is the same
            for all constellation members on each call.
        randomize_lan: Add a random LAN phase shift that is the same for all constellation
            members on each call.
        walker_delta_kwargs: Additional arguments to pass to :class:`walker_delta`.
    """

    def walker_delta_arg_setup(satellites):
        oe_all = walker_delta(
            n_spacecraft=len(satellites),
            n_planes=n_planes,
            altitude=altitude * 1e3,
            inc=inc,
            rel_phasing=rel_phasing,
            **walker_delta_kwargs,
        )

        if randomize_true_anomaly:
            f_offset = np.random.uniform(0, 2 * np.pi)
            for oe in oe_all:
                oe.f = np.mod(oe.f + f_offset, 2 * np.pi)

        if randomize_lan:
            Omega_offset = np.random.uniform(0, 2 * np.pi)
            for oe in oe_all:
                oe.Omega = np.mod(oe.Omega + Omega_offset, 2 * np.pi)

        return {satellite: {"oe": oe} for satellite, oe in zip(satellites, oe_all)}

    return walker_delta_arg_setup


def relative_to_chief(
    chief_name: str,
    chief_orbit: Union[ClassicElements, Callable],
    deputy_relative_state: dict[str, Union[np.ndarray, Callable]],
):
    """Generate a function to generate a satellite state relative to another satellite.

    The output of this function should be used to set the ``sat_arg_randomizer`` of the
    environment.

    Args:
        chief_name: Name of the chief satellite.
        chief_orbit: Orbital elements of the chief satellite, or orbit randomizing function.
        deputy_relative_state: Relative state of the deputy satellites. Should be a dict of
            deputy names and relative states as 6-element vector (rho_H, rho_dot_H) or
            function that returns such a vector.
    """

    def relative_orbit_arg_setup(satellites):
        args = {sat: {} for sat in satellites}

        chief = [satellite for satellite in satellites if satellite.name == chief_name][
            0
        ]
        mu = chief.sat_args_generator["mu"]

        if isinstance(chief_orbit, ClassicElements):
            args[chief]["oe"] = chief_orbit
        else:
            args[chief]["oe"] = chief_orbit()

        rc_N, vc_N = elem2rv(mu, args[chief]["oe"])

        for satellite_name, relative_state in deputy_relative_state.items():
            deputy = [
                satellite
                for satellite in satellites
                if satellite.name == satellite_name
            ][0]

            if not isinstance(relative_state, list) and not isinstance(
                relative_state, np.ndarray
            ):
                relative_state = relative_state()

            rho_H = relative_state[0:3]
            rho_deriv_H = relative_state[3:6]
            rd_N, vd_N = hill2cd(rc_N, vc_N, rho_H, rho_deriv_H)
            args[deputy]["oe"] = rv2elem(mu, rd_N, vd_N)

        return args

    return relative_orbit_arg_setup


class TrajectorySimulator(SimulationBaseClass.SimBaseClass):
    """Class for propagating trajectory using a point mass simulation."""

    def __init__(
        self,
        utc_init: str,
        rN: Optional[Iterable[float]] = None,
        vN: Optional[Iterable[float]] = None,
        oe: Optional[ClassicElements] = None,
        mu: Optional[float] = None,
        dt: float = 30.0,
    ) -> None:
        """Class for propagating trajectory using a point mass simulation.

        Simulated under the effect of Earth's gravity. Returns interpolators for
        position as well as upcoming eclipse predictions. Specify either ``(rN, vN)`` or
        ``(oe, mu)``.

        Args:
            utc_init: Simulation start time.
            rN: Initial position [m]
            vN: Initial velocity [m/s]
            oe: Orbital elements.
            mu: Gravitational parameter
            dt: Simulation timestep.
        """
        super().__init__()
        self.utc_init = utc_init
        self.dt = dt
        rv_specified = rN is not None and vN is not None
        rv_specified_partial = rN is not None or vN is not None
        oe_specified = oe is not None

        if rv_specified and not oe_specified:
            self.rN_init = rN
            self.vN_init = vN
        elif oe_specified and not rv_specified_partial:
            self.rN_init, self.vN_init = elem2rv(mu, oe)
        else:
            raise (
                ValueError(
                    "Orbit is over or underspecified. "
                    + "Provide either (rN, vN) or (oe, mu)"
                )
            )

        self._eclipse_starts: list[float] = []
        self._eclipse_ends: list[float] = []
        self._eclipse_search_time = 0.0

        self._init_simulator()

    def _init_simulator(self) -> None:
        simTaskName = "simTask"
        simProcessName = "simProcess"
        dynProcess = self.CreateNewProcess(simProcessName)
        simulationTimeStep = macros.sec2nano(self.dt)
        dynProcess.addTask(self.CreateNewTask(simTaskName, simulationTimeStep))
        scObject = spacecraft.Spacecraft()
        scObject.ModelTag = "traj-sat"
        self.AddModelToTask(simTaskName, scObject)

        scObject.hub.r_CN_NInit = self.rN_init  # m
        scObject.hub.v_CN_NInit = self.vN_init  # m/s

        # Set up gravity body
        self.gravFactory = simIncludeGravBody.gravBodyFactory()
        planet = self.gravFactory.createEarth()
        self.gravFactory.createSun()
        planet.isCentralBody = True
        planet.useSphericalHarmonicsGravityModel(
            bskPath + "/supportData/LocalGravData/GGM03S.txt", 10
        )
        UTCInit = self.utc_init
        self.gravFactory.createSpiceInterface(
            bskPath + "/supportData/EphemerisData/", UTCInit
        )
        self.gravFactory.spiceObject.zeroBase = "earth"
        self.AddModelToTask(simTaskName, self.gravFactory.spiceObject)
        scObject.gravField.gravBodies = spacecraft.GravBodyVector(
            list(self.gravFactory.gravBodies.values())
        )

        # Set up eclipse
        self.eclipseObject = eclipse.Eclipse()
        self.eclipseObject.addPlanetToModel(
            self.gravFactory.spiceObject.planetStateOutMsgs[0]
        )
        self.eclipseObject.sunInMsg.subscribeTo(
            self.gravFactory.spiceObject.planetStateOutMsgs[1]
        )
        self.AddModelToTask(simTaskName, self.eclipseObject, ModelPriority=988)
        self.eclipseObject.addSpacecraftToModel(scObject.scStateOutMsg)

        # Log outputs
        self.sc_state_log = scObject.scStateOutMsg.recorder()
        self.planet_state_log = self.gravFactory.spiceObject.planetStateOutMsgs[
            0
        ].recorder()
        self.eclipse_log = self.eclipseObject.eclipseOutMsgs[0].recorder()

        self._time_log = []
        self.AddModelToTask(simTaskName, self.sc_state_log)
        self._r_BN_N_log = []
        self.AddModelToTask(simTaskName, self.planet_state_log)
        self._J20002Pfix_log = []
        self.AddModelToTask(simTaskName, self.eclipse_log)
        self._shadowFactor_log = []

        self.InitializeSimulation()

    @property
    def sim_time(self) -> float:
        """Current simulator end time."""
        return macros.NANO2SEC * self.TotalSim.CurrentNanos

    @property
    def times(self) -> np.ndarray:
        """Recorder times in seconds."""
        return macros.NANO2SEC * np.array(self._time_log)

    def extend_to(self, t: float) -> None:
        """Compute the trajectory of the satellite up to t.

        Args:
            t: Computation end [s]
        """
        if t < self.sim_time:
            return
        self.ConfigureStopTime(macros.sec2nano(t))
        self.ExecuteSimulation()
        self._time_log.extend(self.sc_state_log.times())
        self._r_BN_N_log.extend(self.sc_state_log.r_BN_N)
        self._J20002Pfix_log.extend(self.planet_state_log.J20002Pfix)
        self._shadowFactor_log.extend(self.eclipse_log.shadowFactor)
        self.sc_state_log.clear()
        self.planet_state_log.clear()
        self.eclipse_log.clear()

    def _generate_eclipses(self, t: float) -> None:
        self.extend_to(t + self.dt)
        upcoming_times = self.times[self.times > self._eclipse_search_time]
        upcoming_eclipse = (
            np.array(self._shadowFactor_log)[self.times > self._eclipse_search_time] > 0
        ).astype(float)
        for i in np.where(np.diff(upcoming_eclipse) == -1)[0]:
            self._eclipse_starts.append(upcoming_times[i])
        for i in np.where(np.diff(upcoming_eclipse) == 1)[0]:
            self._eclipse_ends.append(upcoming_times[i])

        self._eclipse_search_time = t

    def next_eclipse(self, t: float, max_tries: int = 100) -> tuple[float, float]:
        """Find the soonest eclipse transitions.

        The returned values are not necessarily from the same eclipse event, such as
        when the search start time is in eclipse.

        Args:
            t: Time to start searching [s]
            max_tries: Maximum number of times to search

        Returns:
            eclipse_start: Nearest upcoming eclipse beginning
            eclipse_end:  Nearest upcoming eclipse end
        """
        for i in range(max_tries):
            if any([t_start > t for t_start in self._eclipse_starts]) and any(
                [t_end > t for t_end in self._eclipse_ends]
            ):
                eclipse_start = min(
                    [t_start for t_start in self._eclipse_starts if t_start > t]
                )
                eclipse_end = min([t_end for t_end in self._eclipse_ends if t_end > t])
                return eclipse_start, eclipse_end

            self._generate_eclipses(t + i * self.dt * 10)

        return 1.0, 1.0

    @property
    def r_BN_N(self) -> interp1d:
        """Interpolator for r_BN_N."""
        if self.sim_time < self.dt * 3:
            self.extend_to(self.dt * 3)
        return interp1d(
            self.times,
            self._r_BN_N_log,
            kind="cubic",
            axis=0,
            fill_value="extrapolate",
        )

    @property
    def r_BP_P(self) -> interp1d:
        """Interpolator for r_BP_P."""
        if self.sim_time < self.dt * 3:
            self.extend_to(self.dt * 3)
        return interp1d(
            self.times,
            [
                np.matmul(dcm, pos)
                for dcm, pos in zip(self._J20002Pfix_log, self._r_BN_N_log)
            ],
            kind="cubic",
            axis=0,
            fill_value="extrapolate",
        )

    def __del__(self) -> None:
        """Unload spice kernels when object is deleted."""
        try:
            self.gravFactory.unloadSpiceKernels()
        except AttributeError:
            pass


def lla2ecef(lat: float, long: float, radius: float):
    """Project LLA to Earth Centered, Earth Fixed location.

    Args:
        lat: [deg]
        long: [deg]
        radius: [any]
    """
    lat = np.radians(lat)
    long = np.radians(long)
    return radius * np.array(
        [np.cos(lat) * np.cos(long), np.cos(lat) * np.sin(long), np.sin(lat)]
    )


def rv2HN(r_N: np.ndarray, v_N: np.ndarray):
    """Find the Hill frame rotation matrix from position and velocity.

    Args:
        r_N: Position vector in the inertial frame
        v_N: Velocity vector in the inertial frame
    Returns:
        Hill frame rotation matrix HN
    """
    o_r_N = r_N / np.linalg.norm(r_N)
    # Manually compute h_N = cross(r_N, v_N)
    h_N = np.array(
        [
            r_N[1] * v_N[2] - r_N[2] * v_N[1],
            r_N[2] * v_N[0] - r_N[0] * v_N[2],
            r_N[0] * v_N[1] - r_N[1] * v_N[0],
        ]
    )
    o_h_N = h_N / np.linalg.norm(h_N)
    # Manually compute o_theta_N = cross(o_h_N, o_r_N)
    o_theta_N = np.array(
        [
            o_h_N[1] * o_r_N[2] - o_h_N[2] * o_r_N[1],
            o_h_N[2] * o_r_N[0] - o_h_N[0] * o_r_N[2],
            o_h_N[0] * o_r_N[1] - o_h_N[1] * o_r_N[0],
        ]
    )
    HN = np.array([o_r_N, o_theta_N, o_h_N])
    return HN


def rv2omega(r_N: np.ndarray, v_N: np.ndarray):
    """Find the Hill frame rotation rate from position and velocity.

    Args:
        r_N: Position vector in the inertial frame
        v_N: Velocity vector in the inertial frame
    Returns:
        omega_HN_N: Angular velocity of the Hill frame in the inertial frame
    """
    o_r_N = r_N / np.linalg.norm(r_N)
    h_N = np.cross(r_N, v_N)
    o_h_N = h_N / np.linalg.norm(h_N)
    o_theta_N = np.cross(o_h_N, o_r_N)
    omega_HN_N = o_h_N * np.dot(v_N, o_theta_N) / np.linalg.norm(r_N)
    return omega_HN_N


def cd2hill(
    rc_N: np.ndarray, vc_N: np.ndarray, rd_N: np.ndarray, vd_N: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """Convert from chief and deputy states to Hill frame relative states.

    Args:
        rc_N: Inertial position of chief
        vc_N: Inertial velocity of chief
        rd_N: Inertial position of deputy
        vd_N: Inertial velocity of deputy

    Returns:
        rho_H: Relative position in Hill frame
        rho_deriv_H: Relative velocity in Hill frame
    """
    h_N = np.cross(rc_N, vc_N)
    o_h_N = h_N / np.linalg.norm(h_N)
    ON = rv2HN(rc_N, vc_N)

    f_dot = np.linalg.norm(h_N) / np.linalg.norm(rc_N) ** 2
    omega_ON_N = f_dot * o_h_N

    rho_H = ON @ (rd_N - rc_N)
    rho_deriv_H = ON @ (vd_N - vc_N) - np.cross(ON @ omega_ON_N, rho_H)

    return rho_H, rho_deriv_H


def hill2cd(
    rc_N: np.ndarray, vc_N: np.ndarray, rho_H: np.ndarray, rho_deriv_H: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """Convert from chief and Hill frame relative states to deputy inertial state.

    Args:
        rc_N: Inertial position of chief
        vc_N: Inertial velocity of chief
        rho_H: Relative position in Hill frame
        rho_deriv_H: Relative velocity in Hill frame

    Returns:
        rd_N: Inertial position of deputy
        vd_N: Inertial velocity of deputy
    """
    h_N = np.cross(rc_N, vc_N)
    o_h_N = h_N / np.linalg.norm(h_N)
    ON = rv2HN(rc_N, vc_N)

    f_dot = np.linalg.norm(h_N) / np.linalg.norm(rc_N) ** 2
    omega_ON_N = f_dot * o_h_N

    rd_N = rc_N + ON.T @ rho_H
    vd_N = vc_N + ON.T @ (rho_deriv_H + np.cross(ON @ omega_ON_N, rho_H))

    return rd_N, vd_N


def fibonacci_sphere(n_points):
    """Generate points on a sphere using the Fibonacci sphere method.

    Args:
        n_points: Number of points to generate on the sphere.
    """
    # https://gist.github.com/Seanmatthews/a51ac697db1a4f58a6bca7996d75f68c
    ga = (3 - np.sqrt(5)) * np.pi  # golden angle
    theta = ga * np.arange(n_points)
    z = np.linspace(1 / n_points - 1, 1 - 1 / n_points, n_points)
    radius = np.sqrt(1 - z * z)
    y = radius * np.sin(theta)
    x = radius * np.cos(theta)

    return np.vstack((x, y, z)).T


__doc_title__ = "Orbital"
__all__ = [
    "random_orbit",
    "random_circular_orbit",
    "random_epoch",
    "lla2ecef",
    "elevation",
    "walker_delta",
    "walker_delta_args",
    "relative_to_chief",
    "TrajectorySimulator",
    "rv2HN",
    "rv2omega",
    "cd2hill",
    "hill2cd",
    "fibonacci_sphere",
]



from unittest.mock import MagicMock, patch

import pytest

from bsk_rl.comm import (
    BroadcastCommunication,
    CommunicationMethod,
    FreeCommunication,
    LOSCommunication,
    MultiDegreeCommunication,
    NoCommunication,
)
from bsk_rl.sim.dyn import LOSCommDynModel


@patch.multiple(CommunicationMethod, __abstractmethods__=set())
class TestCommunicationMethod:
    def test_communicate(self):
        mock_sats = [MagicMock(), MagicMock(), MagicMock()]
        mock_sats[0].simulator.sim_time = 0.0
        comms = CommunicationMethod()
        comms.last_communication_time = 0.0
        comms.link_satellites(mock_sats)
        comms.communication_pairs = MagicMock(
            return_value=[(mock_sats[0], mock_sats[1])]
        )
        comms.communicate()
        mock_sats[1].data_store.stage_communicated_data.assert_called_once_with(
            mock_sats[0].data_store.data
        )
        for sat in mock_sats:
            sat.data_store.update_with_communicated_data.assert_called_once()

    def test_min_period_elapsed(self):
        mock_sats = [MagicMock(), MagicMock(), MagicMock()]
        comms = CommunicationMethod(min_period=1.0)
        comms.link_satellites(mock_sats)
        comms.communication_pairs = MagicMock(
            return_value=[(mock_sats[1], mock_sats[0])]
        )
        comms.last_communication_time = 0.0
        mock_sats[0].simulator.sim_time = 2.0
        comms.communicate()
        for sat in mock_sats:
            sat.data_store.update_with_communicated_data.assert_called_once()

    def test_min_period_not_elapsed(self):
        mock_sats = [MagicMock(), MagicMock(), MagicMock()]
        comms = CommunicationMethod(min_period=1.0)
        comms.link_satellites(mock_sats)
        comms.communication_pairs = MagicMock(
            return_value=[(mock_sats[1], mock_sats[0])]
        )
        comms.last_communication_time = 0.0
        mock_sats[0].simulator.sim_time = 0.5
        comms.communicate()
        for sat in mock_sats:
            sat.data_store.update_with_communicated_data.assert_not_called()

    def test_override_communicate_all(self):
        mock_sats = [MagicMock() for i in range(3)]
        comms = FreeCommunication()
        comms.last_communication_time = 0.0
        mock_sats[0].simulator.sim_time = 1.0
        comms.link_satellites(mock_sats)
        comms._communicate_all = MagicMock()
        comms.communicate()
        comms._communicate_all.assert_called_once()

    def test_override_communicate_all_nocall(self):
        mock_sats = [MagicMock() for i in range(3)]
        comms = CommunicationMethod()
        comms.communication_pairs = MagicMock(
            return_value=[(mock_sats[1], mock_sats[0])]
        )
        comms.last_communication_time = 0.0
        mock_sats[0].simulator.sim_time = 1.0
        comms.link_satellites(mock_sats)
        comms._communicate_all = MagicMock()
        comms.communicate()
        comms._communicate_all.assert_not_called()


class TestNoCommunication:
    def test_communicate(self):
        mock_sats = [MagicMock(), MagicMock()]
        mock_sats[0].simulator.sim_time = 0.0
        comms = NoCommunication()
        comms.last_communication_time = 0.0
        comms.link_satellites(mock_sats)
        comms.communicate()
        mock_sats[0].data_store.stage_communicated_data.assert_not_called()


class TestFreeCommunication:
    def test_communication_pairs(self):
        mock_sats = [MagicMock() for i in range(4)]
        comms = FreeCommunication()
        comms.link_satellites(mock_sats)
        pairs = comms.communication_pairs()
        for sat1, sat2 in zip(mock_sats, mock_sats):
            if sat1 != sat2:
                assert (sat1, sat2) in pairs or (sat2, sat1) in pairs


class TestLOSCommunication:
    def test_dyn_model_check_valid(self):
        mock_sats = [MagicMock(dyn_type=LOSCommDynModel) for i in range(3)]
        comms = LOSCommunication()
        comms.link_satellites(mock_sats)

    def test_dyn_model_check_invalid(self):
        mock_sats = [MagicMock(dyn_type="NotLOSComm") for i in range(3)]
        with pytest.raises(Exception):
            comms = LOSCommunication()
            comms.link_satellites(mock_sats)

    class LOSDynMock(MagicMock, LOSCommDynModel):
        pass

    def test_reset(self):
        mock_sats = [
            MagicMock(dyn_type=LOSCommDynModel, dynamics=self.LOSDynMock())
            for i in range(3)
        ]
        comms = LOSCommunication()
        comms.link_satellites(mock_sats)
        comms.reset_post_sim_init()
        for sat1 in mock_sats:
            assert sat1 in comms.los_logs
            for sat2 in mock_sats:
                if sat2 != sat1:
                    assert (
                        comms.los_logs[sat1][sat2]
                        == sat1.dynamics.losComms.accessOutMsgs.__getitem__().recorder()
                    )

    @pytest.mark.parametrize(
        "access1,access2,access",
        [
            ([0, 0], [0, 0], False),
            ([0, 0], [0, 1], True),
            ([1, 0], [0, 0], True),
            ([1, 0], [0, 1], True),
        ],
    )
    def test_communication_pairs(self, access1, access2, access):
        mock_sats = [MagicMock(dyn_type=LOSCommDynModel) for i in range(2)]
        comms = LOSCommunication()
        comms.link_satellites(mock_sats)
        comms.los_logs = {
            mock_sats[0]: {mock_sats[1]: MagicMock(hasAccess=access1)},
            mock_sats[1]: {mock_sats[0]: MagicMock(hasAccess=access2)},
        }
        if access:
            assert len(comms.communication_pairs()) >= 1
        else:
            assert len(comms.communication_pairs()) == 0

    @patch(
        "bsk_rl.comm.CommunicationMethod.communicate",
        MagicMock(),
    )
    def test_communicate(self):
        mock_sats = [MagicMock(dyn_type=LOSCommDynModel) for i in range(2)]
        comms = LOSCommunication()
        comms.link_satellites(mock_sats)
        loggerA = MagicMock()
        loggerB = MagicMock()
        comms.los_logs = {
            mock_sats[0]: {mock_sats[1]: loggerA},
            mock_sats[1]: {mock_sats[0]: loggerB},
        }
        comms.communicate()
        loggerA.clear.assert_called_once()
        loggerB.clear.assert_called_once()


class TestMultiDegreeCommunication:
    @patch("bsk_rl.comm.CommunicationMethod.communication_pairs")
    def test_communication_pairs(self, mock_pairs):
        mock_sats = [MagicMock() for i in range(6)]
        mock_pairs.return_value = [
            (mock_sats[0], mock_sats[1]),
            (mock_sats[1], mock_sats[3]),
            (mock_sats[3], mock_sats[4]),
        ]
        comms = MultiDegreeCommunication()
        comms.link_satellites(mock_sats)
        pairs = comms.communication_pairs()

        assert (mock_sats[0], mock_sats[4]) in pairs
        assert (mock_sats[0], mock_sats[5]) not in pairs


class FreeBroadcast(BroadcastCommunication, FreeCommunication):
    pass


class NoBroadcast(BroadcastCommunication, NoCommunication):
    pass


class TestBroadcastCommunication:
    @pytest.mark.parametrize(
        "spec1", [[], [None, None], [MagicMock(broadcast_pending=False)]]
    )
    @pytest.mark.parametrize("broadcast1", [False, True])
    @pytest.mark.parametrize(
        "spec2", [[], [None, None], [MagicMock(broadcast_pending=False)]]
    )
    @pytest.mark.parametrize("broadcast2", [False, True])
    @pytest.mark.parametrize("comm_type", [FreeBroadcast, NoBroadcast])
    def test_communication_pairs(self, spec1, broadcast1, spec2, broadcast2, comm_type):
        mock_sats = [MagicMock() for i in range(2)]
        mock_sats[0].action_builder.action_spec = spec1
        if broadcast1:
            mock_sats[0].action_builder.action_spec.append(
                MagicMock(broadcast_pending=True)
            )
        mock_sats[1].action_builder.action_spec = spec2
        if broadcast2:
            mock_sats[1].action_builder.action_spec.append(
                MagicMock(broadcast_pending=True)
            )

        comms = comm_type()
        comms.link_satellites(mock_sats)
        pairs = comms.communication_pairs()

        if comm_type is NoBroadcast:
            assert len(pairs) == 0
        else:
            assert len(pairs) == broadcast1 + broadcast2
            if broadcast1:
                assert (mock_sats[0], mock_sats[1]) in pairs
            if broadcast2:
                assert (mock_sats[1], mock_sats[0]) in pairs



"""Basic FSW model for BSK-RL."""

from abc import ABC, abstractmethod
from functools import wraps
from typing import TYPE_CHECKING, Callable, Iterable, Optional
from weakref import proxy

import numpy as np
from Basilisk.architecture import messaging
from Basilisk.fswAlgorithms import (
    attTrackingError,
    hillPoint,
    locationPointing,
    mrpFeedback,
    mrpSteering,
    rateServoFullNonlinear,
    rwMotorTorque,
    thrForceMapping,
    thrMomentumDumping,
    thrMomentumManagement,
)
from Basilisk.utilities import macros as mc

from bsk_rl.sim import dyn
from bsk_rl.utils.functional import (
    AbstractClassProperty,
    check_aliveness_checkers,
    default_args,
)

if TYPE_CHECKING:  # pragma: no cover
    from bsk_rl.sats import Satellite
    from bsk_rl.sim import Simulator
    from bsk_rl.sim.dyn.base import DynamicsModel
    from bsk_rl.sim.world import WorldModel


def action(
    func: Callable[..., None],
) -> Callable[Callable[..., None], Callable[..., None]]:
    """Decorator to reset the satellite software before executing an action.

    Each time an action is called, the FSW tasks and dynamics models call their
    ``reset_for_action`` methods to ensure that the satellite is in a consistent state
    before the action is executed.

    Action functions are typically called by :ref:`bsk_rl.act` to task the satellite.
    """

    @wraps(func)
    def inner(self, *args, **kwargs) -> Callable[..., None]:
        self.fsw_proc.disableAllTasks()
        self._zero_gateway_msgs()
        self.dynamics.reset_for_action()
        for task in self.tasks:
            task.reset_for_action()
        return func(self, *args, **kwargs)

    inner.__doc__ = "*Decorated with* :class:`~bsk_rl.sim.fsw.action`\n\n" + str(
        func.__doc__
    )

    return inner


class FSWModel(ABC):
    """Abstract Basilisk flight software model."""

    @classmethod
    def _requires_dyn(cls) -> list[type["DynamicsModel"]]:
        """Define minimum :class:`DynamicsModel` for compatibility."""
        return []

    def __init__(
        self, satellite: "Satellite", fsw_rate: float, priority: int = 100, **kwargs
    ) -> None:
        """The abstract base flight software model.

        One FSWModel is instantiated for each satellite in the environment each time the
        environment is reset and new simulator is created.

        Args:
            satellite: Satellite modelled by this model
            fsw_rate: [s] Rate of FSW simulation.
            priority: Model priority.
            kwargs: Passed to task creation functions
        """
        self.satellite = satellite
        self.logger = self.satellite.logger.getChild(self.__class__.__name__)

        for required in self._requires_dyn():
            if not issubclass(satellite.dyn_type, required):
                raise TypeError(
                    f"{satellite.dyn_type} must be a subclass of {required} to "
                    + f"use FSW model of type {self.__class__}"
                )

        fsw_proc_name = "FSWProcess" + self.satellite.name
        self.fsw_proc = self.simulator.CreateNewProcess(fsw_proc_name, priority)
        self.fsw_rate = fsw_rate

        self.tasks = self._make_task_list()

        for task in self.tasks:
            task.create_task()

        for task in self.tasks:
            task._create_module_data()

        self._set_messages()

        for task in self.tasks:
            task._setup_fsw_objects(**kwargs)

        self.fsw_proc.disableAllTasks()

    @property
    def simulator(self) -> "Simulator":
        """Reference to the episode simulator."""
        return self.satellite.simulator

    @property
    def world(self) -> "WorldModel":
        """Reference to the episode world model."""
        return self.simulator.world

    @property
    def dynamics(self) -> "DynamicsModel":
        """Reference to the satellite dynamics model for the episode."""
        return self.satellite.dynamics

    def _make_task_list(self) -> list["Task"]:
        return []

    @abstractmethod  # pragma: no cover
    def _set_messages(self) -> None:
        """Message setup after task creation."""
        pass

    def is_alive(self, log_failure=False) -> bool:
        """Check if the FSW model has failed any aliveness requirements.

        Returns:
            ``True`` if the satellite FSW is still alive.
        """
        return check_aliveness_checkers(self, log_failure=log_failure)

    def __del__(self):
        """Log when FSW model is deleted."""
        self.logger.debug("Basilisk FSW deleted")


class Task(ABC):
    """Abstract class for defining FSW tasks."""

    name: str = AbstractClassProperty()

    def __init__(self, fsw: FSWModel, priority: int) -> None:
        """Template class for defining FSW processes.

        Each FSW process has a task associated with it, which handle certain housekeeping
        functions.

        Args:
            fsw: FSW model task contributes to
            priority: Task priority
        """
        self.fsw: FSWModel = proxy(fsw)
        self.priority = priority

    def create_task(self) -> None:
        """Add task to FSW with a unique name."""
        self.fsw.fsw_proc.addTask(
            self.fsw.simulator.CreateNewTask(
                self.name + self.fsw.satellite.name, mc.sec2nano(self.fsw.fsw_rate)
            ),
            taskPriority=self.priority,
        )

    @abstractmethod  # pragma: no cover
    def _create_module_data(self) -> None:
        """Create module data wrappers."""
        pass

    @abstractmethod  # pragma: no cover
    def _setup_fsw_objects(self, **kwargs) -> None:
        """Initialize model parameters with satellite arguments."""
        pass

    def _add_model_to_task(self, module, priority) -> None:
        """Add a model to this task.

        Args:
            module: Basilisk module
            priority: Model priority
        """
        self.fsw.simulator.AddModelToTask(
            self.name + self.fsw.satellite.name,
            module,
            ModelPriority=priority,
        )

    def reset_for_action(self) -> None:
        """Housekeeping for task when a new action is called.

        Disables task by default, can be overridden by subclasses.
        """
        self.fsw.simulator.disableTask(self.name + self.fsw.satellite.name)


class BasicFSWModel(FSWModel):
    """Basic FSW model with minimum necessary Basilisk components."""

    @classmethod
    def _requires_dyn(cls) -> list[type["DynamicsModel"]]:
        return [dyn.BasicDynamicsModel]

    def _make_task_list(self) -> list[Task]:
        return [
            self.SunPointTask(self),
            self.NadirPointTask(self),
            self.RWDesatTask(self),
            self.TrackingErrorTask(self),
            self.MRPControlTask(self),
        ]

    def _set_messages(self) -> None:
        self._set_config_msgs()
        self._set_gateway_msgs()

    def _set_config_msgs(self) -> None:
        self._set_vehicle_config_msg()
        self._set_thrusters_config_msg()
        self._set_rw_config_msg()

    def _set_vehicle_config_msg(self) -> None:
        """Set the vehicle configuration message."""
        # Use the same inertia in the FSW algorithm as in the simulation
        vehicleConfigOut = messaging.VehicleConfigMsgPayload()
        vehicleConfigOut.ISCPntB_B = self.dynamics.I_mat
        self.vcConfigMsg = messaging.VehicleConfigMsg().write(vehicleConfigOut)

    def _set_thrusters_config_msg(self) -> None:
        """Import the thrusters configuration information."""
        self.thrusterConfigMsg = self.dynamics.thrFactory.getConfigMessage()

    def _set_rw_config_msg(self) -> None:
        """Configure RW pyramid exactly as it is in dynamics."""
        self.fswRwConfigMsg = self.dynamics.rwFactory.getConfigMessage()

    def _set_gateway_msgs(self) -> None:
        """Create C-wrapped gateway messages."""
        self.attRefMsg = messaging.AttRefMsg_C()
        self.attGuidMsg = messaging.AttGuidMsg_C()

        self._zero_gateway_msgs()

        # connect gateway FSW effector command msgs with the dynamics
        self.dynamics.rwStateEffector.rwMotorCmdInMsg.subscribeTo(
            self.rwMotorTorque.rwMotorTorqueOutMsg
        )
        self.dynamics.thrusterSet.cmdsInMsg.subscribeTo(
            self.thrDump.thrusterOnTimeOutMsg
        )

    def _zero_gateway_msgs(self) -> None:
        """Zero all the FSW gateway message payloads."""
        self.attRefMsg.write(messaging.AttRefMsgPayload())
        self.attGuidMsg.write(messaging.AttGuidMsgPayload())

    @action
    def action_drift(self) -> None:
        """Disable all tasks and do nothing."""
        self.simulator.disableTask(
            BasicFSWModel.MRPControlTask.name + self.satellite.name
        )

    class SunPointTask(Task):
        """Task to generate sun-pointing reference."""

        name = "sunPointTask"

        def __init__(self, fsw, priority=99) -> None:  # noqa: D107
            """Task to generate a sun-pointing reference."""
            super().__init__(fsw, priority)

        def _create_module_data(self) -> None:
            self.sunPoint = self.fsw.sunPoint = locationPointing.locationPointing()
            self.sunPoint.ModelTag = "sunPoint"

        def _setup_fsw_objects(self, **kwargs) -> None:
            """Configure the solar array sun-pointing task."""
            self.setup_sun_pointing(**kwargs)

        def setup_sun_pointing(self, nHat_B: Iterable[float], **kwargs) -> None:
            """Configure the solar array sun-pointing task.

            Args:
                nHat_B: Solar array normal vector.
                kwargs: Passed to other setup functions.
            """
            self.sunPoint.pHat_B = nHat_B
            self.sunPoint.scAttInMsg.subscribeTo(
                self.fsw.dynamics.simpleNavObject.attOutMsg
            )
            self.sunPoint.scTransInMsg.subscribeTo(
                self.fsw.dynamics.simpleNavObject.transOutMsg
            )
            self.sunPoint.celBodyInMsg.subscribeTo(
                self.fsw.world.ephemConverter.ephemOutMsgs[self.fsw.world.sun_index]
            )
            self.sunPoint.useBoresightRateDamping = 1
            messaging.AttGuidMsg_C_addAuthor(
                self.sunPoint.attGuidOutMsg, self.fsw.attGuidMsg
            )

            self._add_model_to_task(self.sunPoint, priority=1200)

    @action
    def action_charge(self) -> None:
        """Charge battery by pointing the solar panels at the sun."""
        self.sunPoint.Reset(self.simulator.sim_time_ns)
        self.simulator.enableTask(self.SunPointTask.name + self.satellite.name)

    class NadirPointTask(Task):
        """Task to generate nadir-pointing reference."""

        name = "nadirPointTask"

        def __init__(self, fsw, priority=98) -> None:  # noqa: D107
            """Task to generate nadir-pointing reference."""
            super().__init__(fsw, priority)

        def _create_module_data(self) -> None:
            self.hillPoint = self.fsw.hillPoint = hillPoint.hillPoint()
            self.hillPoint.ModelTag = "hillPoint"

        def _setup_fsw_objects(self, **kwargs) -> None:
            """Configure the nadir-pointing task.

            Args:
                kwargs: Passed to other setup functions.
            """
            self.hillPoint.transNavInMsg.subscribeTo(
                self.fsw.dynamics.simpleNavObject.transOutMsg
            )
            self.hillPoint.celBodyInMsg.subscribeTo(
                self.fsw.world.ephemConverter.ephemOutMsgs[self.fsw.world.body_index]
            )
            messaging.AttRefMsg_C_addAuthor(
                self.hillPoint.attRefOutMsg, self.fsw.attRefMsg
            )

            self._add_model_to_task(self.hillPoint, priority=1199)

    class RWDesatTask(Task):
        """Task to desaturate reaction wheels."""

        name = "rwDesatTask"

        def __init__(self, fsw, priority=97) -> None:  # noqa: D107
            """Task to desaturate reaction wheels using thrusters."""
            super().__init__(fsw, priority)

        def _create_module_data(self) -> None:
            """Set up momentum dumping and thruster control."""
            # Momentum dumping configuration
            self.thrDesatControl = self.fsw.thrDesatControl = (
                thrMomentumManagement.thrMomentumManagement()
            )
            self.thrDesatControl.ModelTag = "thrMomentumManagement"

            self.thrDump = self.fsw.thrDump = thrMomentumDumping.thrMomentumDumping()
            self.thrDump.ModelTag = "thrDump"

            # Thruster force mapping configuration
            self.thrForceMapping = self.fsw.thrForceMapping = (
                thrForceMapping.thrForceMapping()
            )
            self.thrForceMapping.ModelTag = "thrForceMapping"

        def _setup_fsw_objects(self, **kwargs) -> None:
            """Set up thrusters and momentum dumping.

            Args:
                kwargs: Passed to other setup functions.
            """
            self.setup_thruster_mapping(**kwargs)
            self.setup_momentum_dumping(**kwargs)

        @default_args(controlAxes_B=[1, 0, 0, 0, 1, 0, 0, 0, 1], thrForceSign=+1)
        def setup_thruster_mapping(
            self, controlAxes_B: Iterable[float], thrForceSign: int, **kwargs
        ) -> None:
            """Configure the thruster mapping.

            Args:
                controlAxes_B: Control unit axes.
                thrForceSign: Flag indicating if pos (+1) or negative (-1) thruster
                    solutions are found.
                kwargs: Passed to other setup functions.
            """
            self.thrForceMapping.cmdTorqueInMsg.subscribeTo(
                self.thrDesatControl.deltaHOutMsg
            )
            self.thrForceMapping.thrConfigInMsg.subscribeTo(self.fsw.thrusterConfigMsg)
            self.thrForceMapping.vehConfigInMsg.subscribeTo(self.fsw.vcConfigMsg)
            self.thrForceMapping.controlAxes_B = controlAxes_B
            self.thrForceMapping.thrForceSign = thrForceSign
            self.thrForceMapping.angErrThresh = 3.15

            self._add_model_to_task(self.thrForceMapping, priority=1192)

        @default_args(
            hs_min=0.0,
            maxCounterValue=4,
            thrMinFireTime=0.02,
            desatAttitude="sun",
        )
        def setup_momentum_dumping(
            self,
            hs_min: float,
            maxCounterValue: int,
            thrMinFireTime: float,
            desatAttitude: Optional[str],
            **kwargs,
        ) -> None:
            """Configure the momentum dumping algorithm.

            Args:
                hs_min: [N*m*s] Minimum RW cluster momentum for dumping.
                maxCounterValue: Control periods between firing thrusters.
                thrMinFireTime: [s] Minimum thruster firing time.
                desatAttitude: Direction to point while desaturating:

                    * ``"sun"`` points panels at sun
                    * ``"nadir"`` points instrument nadir
                    * ``None`` disables attitude control.

                kwargs: Passed to other setup functions.
            """
            self.fsw.desatAttitude = desatAttitude
            self.thrDesatControl.hs_min = hs_min  # Nms
            self.thrDesatControl.rwSpeedsInMsg.subscribeTo(
                self.fsw.dynamics.rwStateEffector.rwSpeedOutMsg
            )
            self.thrDesatControl.rwConfigDataInMsg.subscribeTo(self.fsw.fswRwConfigMsg)

            self.thrDump.deltaHInMsg.subscribeTo(self.thrDesatControl.deltaHOutMsg)
            self.thrDump.thrusterImpulseInMsg.subscribeTo(
                self.thrForceMapping.thrForceCmdOutMsg
            )
            self.thrDump.thrusterConfInMsg.subscribeTo(self.fsw.thrusterConfigMsg)
            self.thrDump.maxCounterValue = maxCounterValue
            self.thrDump.thrMinFireTime = thrMinFireTime

            self._add_model_to_task(self.thrDesatControl, priority=1193)
            self._add_model_to_task(self.thrDump, priority=1191)

        def reset_for_action(self) -> None:
            """Disable power draw for thrusters when a new action is selected."""
            super().reset_for_action()
            self.fsw.dynamics.thrusterPowerSink.powerStatus = 0

    @action
    def action_desat(self) -> None:
        """Charge while desaturating reaction wheels.

        This action maneuvers the satellite into ``desatAttitude``, turns on the thruster
        power sink, and enables the desaturation tasks. This action typically needs to be
        called multiple times to fully desaturate the wheels.
        """
        self.trackingError.Reset(self.simulator.sim_time_ns)
        self.thrDesatControl.Reset(self.simulator.sim_time_ns)
        self.thrDump.Reset(self.simulator.sim_time_ns)
        self.dynamics.thrusterPowerSink.powerStatus = 1
        self.simulator.enableTask(self.RWDesatTask.name + self.satellite.name)
        if self.desatAttitude == "sun":
            self.sunPoint.Reset(self.simulator.sim_time_ns)
            self.simulator.enableTask(self.SunPointTask.name + self.satellite.name)
        elif self.desatAttitude == "nadir":
            self.hillPoint.Reset(self.simulator.sim_time_ns)
            self.simulator.enableTask(
                BasicFSWModel.NadirPointTask.name + self.satellite.name
            )
        elif self.desatAttitude is None:
            pass
        else:
            raise ValueError(f"{self.desatAttitude} not a valid desatAttitude")
        self.simulator.enableTask(self.TrackingErrorTask.name + self.satellite.name)

    class TrackingErrorTask(Task):
        """Task to convert an attitude reference to guidance."""

        name = "trackingErrTask"

        def __init__(self, fsw, priority=90) -> None:  # noqa: D107
            """Task to convert an attitude reference to guidance."""
            super().__init__(fsw, priority)

        def _create_module_data(self) -> None:
            self.trackingError = self.fsw.trackingError = (
                attTrackingError.attTrackingError()
            )
            self.trackingError.ModelTag = "trackingError"

        def _setup_fsw_objects(self, **kwargs) -> None:
            self.trackingError.attNavInMsg.subscribeTo(
                self.fsw.dynamics.simpleNavObject.attOutMsg
            )
            self.trackingError.attRefInMsg.subscribeTo(self.fsw.attRefMsg)
            messaging.AttGuidMsg_C_addAuthor(
                self.trackingError.attGuidOutMsg, self.fsw.attGuidMsg
            )

            self._add_model_to_task(self.trackingError, priority=1197)

    class MRPControlTask(Task):
        """Task to control the satellite attitude using reaction wheels."""

        name = "mrpControlTask"

        def __init__(self, fsw, priority=80) -> None:  # noqa: D107
            """Task to control the satellite with reaction wheels."""
            super().__init__(fsw, priority)

        def _create_module_data(self) -> None:
            # Attitude controller configuration
            self.mrpFeedbackControl = self.fsw.mrpFeedbackControl = (
                mrpFeedback.mrpFeedback()
            )
            self.mrpFeedbackControl.ModelTag = "mrpFeedbackControl"

            # add module that maps the Lr control torque into the RW motor torques
            self.rwMotorTorque = self.fsw.rwMotorTorque = rwMotorTorque.rwMotorTorque()
            self.rwMotorTorque.ModelTag = "rwMotorTorque"

        def _setup_fsw_objects(self, **kwargs) -> None:
            self.setup_mrp_feedback_rwa(**kwargs)
            self.setup_rw_motor_torque(**kwargs)

        @default_args(K=7.0, Ki=-1, P=35.0)
        def setup_mrp_feedback_rwa(
            self, K: float, Ki: float, P: float, **kwargs
        ) -> None:
            """Set the MRP feedback control properties.

            Args:
                K: Proportional gain.
                Ki: Integral gain.
                P: Derivative gain.
                kwargs: Passed to other setup functions.
            """
            self.mrpFeedbackControl.guidInMsg.subscribeTo(self.fsw.attGuidMsg)
            self.mrpFeedbackControl.vehConfigInMsg.subscribeTo(self.fsw.vcConfigMsg)
            self.mrpFeedbackControl.K = K
            self.mrpFeedbackControl.Ki = Ki
            self.mrpFeedbackControl.P = P
            self.mrpFeedbackControl.integralLimit = (
                2.0 / self.mrpFeedbackControl.Ki * 0.1
            )

            self._add_model_to_task(self.mrpFeedbackControl, priority=1196)

        def setup_rw_motor_torque(
            self, controlAxes_B: Iterable[float], **kwargs
        ) -> None:
            """Set parameters for finding motor torque from the control law.

            Args:
                controlAxes_B: Control unit axes.
                kwargs: Passed to other setup functions.
            """
            self.rwMotorTorque.rwParamsInMsg.subscribeTo(self.fsw.fswRwConfigMsg)
            self.rwMotorTorque.vehControlInMsg.subscribeTo(
                self.mrpFeedbackControl.cmdTorqueOutMsg
            )
            self.rwMotorTorque.controlAxes_B = controlAxes_B

            self._add_model_to_task(self.rwMotorTorque, priority=1195)

        def reset_for_action(self) -> None:
            """MRP control is enabled by default for all tasks."""
            self.fsw.simulator.enableTask(self.name + self.fsw.satellite.name)


class SteeringFSWModel(BasicFSWModel):
    """FSW extending MRP control to use MRP steering instead of MRP feedback."""

    def __init__(self, *args, **kwargs) -> None:
        """FSW extending attitude control to use MRP steering instead of MRP feedback.

        This class replaces the simple attitude feedback control law with a more
        sophisticated `MRP steering control law <https://hanspeterschaub.info/basilisk/Documentation/fswAlgorithms/attControl/mrpSteering/mrpSteering.html>`_.
        """
        super().__init__(*args, **kwargs)

    class MRPControlTask(Task):
        """Task that uses MRP steering to control reaction wheels."""

        name = "mrpControlTask"

        def __init__(self, fsw, priority=80) -> None:  # noqa: D107
            """Task to control the satellite with reaction wheels."""
            super().__init__(fsw, priority)

        def _create_module_data(self) -> None:
            # Attitude controller configuration
            self.mrpSteeringControl = self.fsw.mrpSteeringControl = (
                mrpSteering.mrpSteering()
            )
            self.mrpSteeringControl.ModelTag = "mrpSteeringControl"

            # Rate servo
            self.servo = self.fsw.servo = (
                rateServoFullNonlinear.rateServoFullNonlinear()
            )
            self.servo.ModelTag = "rateServo"

            # add module that maps the Lr control torque into the RW motor torques
            self.rwMotorTorque = self.fsw.rwMotorTorque = rwMotorTorque.rwMotorTorque()
            self.rwMotorTorque.ModelTag = "rwMotorTorque"

        def _setup_fsw_objects(self, **kwargs) -> None:
            self.setup_mrp_steering_rwa(**kwargs)
            self.setup_rw_motor_torque(**kwargs)

        @default_args(
            K1=0.25, K3=3.0, omega_max=np.radians(3), servo_Ki=5.0, servo_P=150
        )
        def setup_mrp_steering_rwa(
            self,
            K1: float,
            K3: float,
            omega_max: float,
            servo_Ki: float,
            servo_P: float,
            **kwargs,
        ) -> None:
            """Define the control properties.

            Args:
                K1: MRP steering gain.
                K3: MRP steering gain.
                omega_max: [rad/s] Maximum targetable spacecraft body rate.
                servo_Ki: Servo gain.
                servo_P: Servo gain.
                kwargs: Passed to other setup functions.
            """
            self.mrpSteeringControl.guidInMsg.subscribeTo(self.fsw.attGuidMsg)
            self.mrpSteeringControl.K1 = K1
            self.mrpSteeringControl.K3 = K3
            self.mrpSteeringControl.omega_max = omega_max
            self.mrpSteeringControl.ignoreOuterLoopFeedforward = False

            self.servo.Ki = servo_Ki
            self.servo.P = servo_P
            self.servo.integralLimit = 2.0 / self.servo.Ki * 0.1
            self.servo.knownTorquePntB_B = [0.0, 0.0, 0.0]
            self.servo.guidInMsg.subscribeTo(self.fsw.attGuidMsg)
            self.servo.vehConfigInMsg.subscribeTo(self.fsw.vcConfigMsg)
            self.servo.rwParamsInMsg.subscribeTo(self.fsw.fswRwConfigMsg)
            self.servo.rwSpeedsInMsg.subscribeTo(
                self.fsw.dynamics.rwStateEffector.rwSpeedOutMsg
            )
            self.servo.rateSteeringInMsg.subscribeTo(
                self.mrpSteeringControl.rateCmdOutMsg
            )

            self._add_model_to_task(self.mrpSteeringControl, priority=1196)
            self._add_model_to_task(self.servo, priority=1195)

        def setup_rw_motor_torque(
            self, controlAxes_B: Iterable[float], **kwargs
        ) -> None:
            """Define the motor torque from the control law.

            Args:
                controlAxes_B: Control unit axes.
                kwargs: Passed to other setup functions.
            """
            self.rwMotorTorque.rwParamsInMsg.subscribeTo(self.fsw.fswRwConfigMsg)
            self.rwMotorTorque.vehControlInMsg.subscribeTo(self.servo.cmdTorqueOutMsg)
            self.rwMotorTorque.controlAxes_B = controlAxes_B

            self._add_model_to_task(self.rwMotorTorque, priority=1194)

        def reset_for_action(self) -> None:
            """Keep MRP control enabled on action calls."""
            self.fsw.simulator.enableTask(self.name + self.fsw.satellite.name)


__doc_title__ = "FSW Base"
__all__ = [
    "action",
    "FSWModel",
    "Task",
    "BasicFSWModel",
    "SteeringFSWModel",
]



"""Continuous actions set satellite behavior based on some continuous value."""

import logging
from abc import abstractmethod
from typing import TYPE_CHECKING, Optional

import numpy as np
from gymnasium import spaces

from bsk_rl.act.actions import Action, ActionBuilder

if TYPE_CHECKING:  # pragma: no cover
    from bsk_rl.sats import Satellite

logger = logging.getLogger(__name__)


class ContinuousActionBuilder(ActionBuilder):
    def __init__(self, satellite: "Satellite") -> None:
        """Processes actions for a continuous action space.

        Args:
            satellite: Satellite to create actions for.
        """
        self.action_spec: list[ContinuousAction]
        super().__init__(satellite)
        assert len(self.action_spec) == 1, "Only one continuous action is supported."

    @property
    def _action(self) -> "ContinuousAction":
        return self.action_spec[0]

    @property
    def action_space(self) -> spaces.Box:
        """Continuous action space."""
        return self._action.space

    @property
    def action_description(self) -> list[str]:
        """Return a human-readable description of the continuous action space."""
        return self._action.action_description()

    def set_action(self, action: np.ndarray) -> None:
        """Activate the action by setting the continuous value."""
        self._action.set_action(action)


class ContinuousAction(Action):
    builder_type = ContinuousActionBuilder

    def __init__(self, name: str = "discrete_act") -> None:
        """Base class for actions with a continuous action space.

        Args:
            name: Name of the action.
        """
        super().__init__(name=name)

    @property
    @abstractmethod
    def space(self) -> spaces.Box:
        """Return the action space."""
        pass

    @property
    @abstractmethod
    def action_description(self) -> list[str]:
        """Return a description of the action space."""
        pass

    @abstractmethod
    def set_action(self, action: np.ndarray) -> None:
        """Activate an action by a continuous value."""
        pass


class ImpulsiveThrust(ContinuousAction):
    def __init__(
        self,
        name: str = "thrust_act",
        max_dv: float = float("inf"),
        max_drift_duration: float = float("inf"),
        fsw_action: Optional[str] = None,
    ) -> None:
        """Perform an impulsive thrust and drift for some duration.

        Args:
            name: Name of the action.
            max_dv: Maximum delta-V that can be applied. [m/s]
            max_drift_duration: Maximum duration to drift after applying the delta-V. [s]
            fsw_action: Name of the FSW action to activate during the drift period.
        """
        super().__init__(name)
        self.max_dv = max_dv
        self.max_drift_duration = max_drift_duration
        self.fsw_action = fsw_action

    @property
    def space(self) -> spaces.Box:
        """Return the action space."""
        return spaces.Box(
            low=np.array(
                [-self.max_dv, -self.max_dv, -self.max_dv, 2 * self.simulator.sim_rate]
            ),
            high=np.array(
                [self.max_dv, self.max_dv, self.max_dv, self.max_drift_duration]
            ),
            shape=(4,),
            dtype=np.float32,
        )

    @property
    def action_description(self) -> list[str]:
        """Description of the continuous action space."""
        return ["dV_N_x", "dV_N_y", "dV_N_z", "duration"]

    def set_action(self, action: np.ndarray) -> None:
        """Thrust the satellite with a given inertial delta-V and drift for some duration.

        Args:
            action: vector of [dV_N_x, dV_N_y, dV_N_z, duration] in [m/s] and [s]
        """
        assert len(action) == 4, "Action must have 4 elements."

        # Restrict to maximum delta-V
        dv_N = action[0:3]
        if np.linalg.norm(dv_N) > self.max_dv:
            self.satellite.logger.info(
                f"Thrust clamped from {np.linalg.norm(dv_N)} m/s to {self.max_dv} m/s."
            )
            dv_N = dv_N / np.linalg.norm(dv_N) * self.max_dv

        # Restrict to duration limits
        dt_desired = action[3]
        dt = np.clip(
            dt_desired, 2 * self.satellite.simulator.sim_rate, self.max_drift_duration
        )
        if dt != dt_desired:
            self.satellite.logger.warning(
                f"Requested drift duration {dt_desired} out of range. Clamping to {dt}."
            )

        self.satellite.logger.info(
            f"Thrusting with inertial dV {dv_N} with {dt} second drift."
        )
        self.satellite.fsw.action_impulsive_thrust(dv_N)
        self.satellite.update_timed_terminal_event(
            self.satellite.simulator.sim_time + dt
        )

        # Activate the FSW action for the drift period
        # TODO: should wait until action_impulsive_thrust is done if not immediate
        if self.fsw_action is not None:
            getattr(self.satellite.fsw, self.fsw_action)()
            self.satellite.logger.info(f"FSW action {self.fsw_action} activated.")


class ImpulsiveThrustHill(ImpulsiveThrust):
    def __init__(self, chief_name, *args, **kwargs):
        """Impulsive thrusts in the Hill frame.

        Args:
            chief_name: Chief to use for Hill frame.
            *args: Passed to ``ImpulsiveThrust``.
            **kwargs: Passed to ``ImpulsiveThrust``.
        """
        self.chief_name = chief_name
        super().__init__(*args, **kwargs)

    def reset_post_sim_init(self) -> None:
        """Connect to the chief satellite.

        :meta private:
        """
        self.chief = self.satellite.simulator.get_satellite(self.chief_name)

    def set_action(self, action: np.ndarray) -> None:
        """Activate the action by setting the continuous value."""
        dv_H = action[0:3]
        dt = action[3]

        NH = self.chief.dynamics.HN.T
        dv_N = NH @ dv_H

        super().set_action(np.concatenate((dv_N, [dt])))



from unittest.mock import MagicMock, Mock

import numpy as np
import pytest
from gymnasium import spaces

from bsk_rl import obs
from bsk_rl.obs.observations import ObservationBuilder, _target_angle


class TestObservationBuilder:
    def test_init(self):
        names = ["obs_A", "obs_B", "obs_A"]
        observation_spec = [MagicMock() for _ in names]
        for os, name in zip(observation_spec, names):
            os.name = name

        ob = ObservationBuilder(satellite=MagicMock(observation_spec=observation_spec))
        assert len(ob.observation_spec) == 3
        assert ob.observation_spec[2].name == "obs_A_2"

    def test_reset_post_sim_init(self):
        ob = ObservationBuilder(
            satellite=MagicMock(observation_spec=[MagicMock() for _ in range(3)])
        )
        for os in ob.observation_spec:
            os.link_satellite.assert_called_once()

    def make_mocked_observation(self, obs_type=np.ndarray):
        sat = MagicMock(simulator=MagicMock(sim_time=0.0))
        observation_spec = [
            MagicMock(get_obs=MagicMock(return_value=i)) for i in range(3)
        ]
        for i, os in enumerate(observation_spec):
            os.name = f"obs_{i}"
        sat.observation_spec = observation_spec
        ob = ObservationBuilder(sat, obs_type=obs_type)
        ob.simulator = sat.simulator
        return ob

    def test_get_obs_np(self):
        ob = self.make_mocked_observation(obs_type=np.ndarray)
        assert np.isclose(ob.get_obs(), np.array([0, 1, 2])).all()

    def test_get_obs_list(self):
        ob = self.make_mocked_observation(obs_type=list)
        assert ob.get_obs() == [0, 1, 2]

    def test_get_obs_dict(self):
        ob = self.make_mocked_observation(obs_type=dict)
        assert ob.get_obs() == {"obs_0": 0, "obs_1": 1, "obs_2": 2}

    def test_obs_array_keys(self):
        ob = self.make_mocked_observation()
        assert ob.obs_array_keys() == ["obs_0", "obs_1", "obs_2"]

    def test_obs_cache(self):
        ob = self.make_mocked_observation()
        ob.get_obs()
        # Don't step time forward, expect cached value
        ob.observation_spec[0].get_obs.return_value = 1
        assert ob.get_obs()[0] == 0
        # Step time forward, expect new value
        ob.satellite.simulator.sim_time = 1.0
        assert ob.get_obs()[0] == 1

    @pytest.mark.parametrize(
        "observation,dtype,space",
        [
            (
                np.array([1]),
                np.float32,
                spaces.Box(low=-1e16, high=1e16, shape=(1,), dtype=np.float32),
            ),
            (
                np.array([1, 2]),
                np.float64,
                spaces.Box(low=-1e16, high=1e16, shape=(2,), dtype=np.float64),
            ),
            (
                {"a": 1, "b": {"c": 1}},
                np.float32,
                spaces.Dict(
                    {
                        "a": spaces.Box(
                            low=-1e16, high=1e16, shape=(1,), dtype=np.float32
                        ),
                        "b": spaces.Dict(
                            {
                                "c": spaces.Box(
                                    low=-1e16, high=1e16, shape=(1,), dtype=np.float32
                                )
                            }
                        ),
                    }
                ),
            ),
        ],
    )
    def test_obs_space(self, observation, dtype, space):
        ob = ObservationBuilder(MagicMock(), dtype=dtype)
        ob.get_obs = MagicMock(return_value=observation)
        assert ob.observation_space == space


class TestSatProperties:
    def test_init(self):
        ob = obs.SatProperties(
            dict(prop="hello", module="dynamics", name="hello_prop"),
            dict(prop="world", norm=2.0),
        )
        assert ob.obs_properties[0]["norm"] == 1.0
        assert ob.obs_properties[1]["name"] == "world_normd"

    def test_id_property(self):
        ob = obs.SatProperties(dict(prop="world"))
        ob.satellite = MagicMock(dynamics=Mock(), fsw=Mock(world=1.0))
        del ob.satellite.dynamics.world
        ob.reset_post_sim_init()
        assert ob.obs_properties[0]["module"] == "fsw"

    def test_get_obs(self):
        ob = obs.SatProperties(
            dict(prop="hello", module="dynamics", name="hello_prop", norm=2.0),
        )
        ob.satellite = MagicMock(dynamics=MagicMock(hello=6.0))
        assert ob.get_obs() == {"hello_prop": 3.0}

    def test_get_obs_fn(self):
        mock_fn = MagicMock(return_value=3.0)
        ob = obs.SatProperties(
            dict(prop="hello", fn=mock_fn),
        )
        ob.satellite = MagicMock()
        assert ob.get_obs() == {"hello": 3.0}
        mock_fn.assert_called_once_with(ob.satellite)


class TestTime:
    def test_detect_norm(self):
        ob = obs.Time()
        ob.simulator = MagicMock(sim_time=10.0, time_limit=100.0)
        ob.reset_post_sim_init()
        assert ob.get_obs() == 0.1

    def test_manual_norm(self):
        ob = obs.Time(norm=100.0)
        ob.simulator = MagicMock(sim_time=10.0, time_limit=300.0)
        ob.reset_post_sim_init()
        assert ob.get_obs() == 0.1

    def test_time_remaining(self):
        ob = obs.Time(observe_time_remaining=True)
        ob.simulator = MagicMock(sim_time=10.0, time_limit=100.0)
        ob.reset_post_sim_init()
        assert ob.get_obs() == 0.9


class TestOpportunityProperties:
    def test_fns(self):
        fns = obs.OpportunityProperties._fn_map
        sat = MagicMock(simulator=MagicMock(sim_time=10.0))
        opp = dict(
            type="target",
            object=MagicMock(priority=1.0),
            window=[20.0, 30.0],
            r_LP_P=1.0,
        )
        assert fns["priority"](sat, opp) == 1.0
        assert fns["r_LP_P"](sat, opp) == 1.0
        assert fns["opportunity_open"](sat, opp) == 10.0
        assert fns["opportunity_close"](sat, opp) == 20.0
        assert fns["opportunity_mid"](sat, opp) == 15.0

    def test_target_angle(self):
        sat = MagicMock(
            dynamics=MagicMock(r_BN_P=np.array([1.0, 0.0, 0.0])),
            fsw=MagicMock(c_hat_P=np.array([0.0, 1.0, 0.0])),
        )
        opp = dict(r_LP_P=np.array([0.0, 0.0, 0.0]))
        assert np.isclose(_target_angle(sat, opp), np.pi / 2)

    def test_init(self):
        ob = obs.OpportunityProperties(
            dict(prop="r_LP_P", norm=2.0),
            dict(
                prop="double_priority", fn=lambda sat, opp: opp["target"].priority * 2.0
            ),
            n_ahead_observe=2,
        )
        assert ob.target_properties[0]["fn"]
        assert ob.target_properties[0]["name"] == "r_LP_P_normd"
        assert ob.target_properties[1]["norm"] == 1.0

    def test_get_obs(self):
        ob = obs.OpportunityProperties(
            dict(prop="priority", norm=2.0),
            dict(
                prop="double_priority", fn=lambda sat, opp: opp["object"].priority * 2.0
            ),
            n_ahead_observe=2,
        )
        ob.satellite = MagicMock()
        ob.satellite.find_next_opportunities.return_value = [
            {"object": MagicMock(priority=1.0), "type": "target"}
        ] * ob.n_ahead_observe
        assert ob.get_obs() == {
            "target_0": {"priority_normd": 0.5, "double_priority": 2.0},
            "target_1": {"priority_normd": 0.5, "double_priority": 2.0},
        }

    def test_init_bad(self):
        with pytest.raises(ValueError):
            obs.OpportunityProperties(dict(prop="not_a_prop"), n_ahead_observe=2)


class TestEclipse:
    def test_obs(self):
        ob = obs.Eclipse(norm=100.0)
        ob.simulator = MagicMock(sim_time=10.0)
        ob.satellite = MagicMock()
        ob.satellite.trajectory.next_eclipse.return_value = (20.0, 30.0)
        assert ob.get_obs() == [0.1, 0.2]


class TestResourceRewardWeight:
    def test_obs(self):
        ob = obs.ResourceRewardWeight()
        ob.reset_overwrite_previous()
        ob.weight_vector.append(1.0)
        assert ob.get_obs() == [1.0]



import gymnasium as gym
import numpy as np
from gymnasium import spaces

from bsk_rl import act, data, obs, sats
from bsk_rl.scene import UniformTargets
from bsk_rl.utils.orbital import random_orbit


class DoNothingSatellite(sats.ImagingSatellite):
    observation_spec = [obs.Time()]
    action_spec = [act.Drift()]


class TestSatelliteTasking:
    env = gym.make(
        "SatelliteTasking-v1",
        satellite=DoNothingSatellite(
            "Sputnik",
            sat_args=DoNothingSatellite.default_sat_args(oe=random_orbit),
        ),
        scenario=UniformTargets(n_targets=0),
        rewarder=data.NoReward(),
        sim_rate=1.0,
        max_step_duration=10.0,
        time_limit=100.0,
        disable_env_checker=True,
        dtype=np.float32,
    )

    def test_reset(self):
        observation, info = self.env.reset()
        assert (observation == np.array([0.0])).all()

    def test_action_space(self):
        assert self.env.action_space == spaces.Discrete(1)

    def test_observation_space(self):
        assert self.env.observation_space == spaces.Box(
            -1e16, 1e16, (1,), dtype=np.float32
        )

    def test_step(self):
        observation, reward, terminated, truncated, info = self.env.step(0)
        assert (observation == np.array([0.1])).all()

    def test_truncate(self):
        terminated = truncated = False
        while not (terminated or truncated):
            observation, reward, terminated, truncated, info = self.env.step(0)
        assert truncated
        assert self.env.unwrapped.simulator.sim_time == 100.0

    def test_repeatable(self):
        self.env.reset(seed=0)
        world_args_old = self.env.unwrapped.world_args
        sat_args_old = self.env.unwrapped.satellite.sat_args
        self.env.reset(seed=0)
        assert self.env.unwrapped.world_args == world_args_old
        for val, val_old in zip(
            self.env.unwrapped.satellite.sat_args.values(), sat_args_old.values()
        ):
            if (
                isinstance(val, (np.ndarray, list, type(None)))
                or np.issubdtype(type(val), np.integer)
                or np.issubdtype(type(val), float)
            ):
                assert np.all(val == val_old)


class TestSingleSatelliteDeath:
    env = gym.make(
        "SatelliteTasking-v1",
        satellite=DoNothingSatellite(
            "Skydiver",
            sat_args=DoNothingSatellite.default_sat_args(
                rN=[0, 0, 7e6], vN=[0, 0, -100.0], oe=None
            ),
        ),
        scenario=UniformTargets(n_targets=0),
        rewarder=data.NoReward(),
        sim_rate=1.0,
        time_limit=1000.0,
        failure_penalty=-1000,
        disable_env_checker=True,
    )

    def test_fail(self):
        observation, info = self.env.reset()
        observation, reward, terminated, truncated, info = self.env.step(0)
        assert terminated
        assert reward == -1000

    def test_close(self):
        self.env.close()


class TestGeneralSatelliteTasking:
    env = gym.make(
        "GeneralSatelliteTasking-v1",
        satellites=[
            DoNothingSatellite(
                "Sentinel-2A",
                sat_args=DoNothingSatellite.default_sat_args(oe=random_orbit),
            ),
            DoNothingSatellite(
                "Sentinel-2B",
                sat_args=DoNothingSatellite.default_sat_args(oe=random_orbit),
            ),
        ],
        scenario=UniformTargets(n_targets=0),
        rewarder=data.NoReward(),
        sim_rate=1.0,
        max_step_duration=10.0,
        time_limit=100.0,
        disable_env_checker=True,
        dtype=np.float32,
    )

    def test_reset(self):
        observation, info = self.env.reset()
        assert (observation == np.array([0.0])).all()

    def test_action_space(self):
        assert self.env.action_space == spaces.Tuple(
            (spaces.Discrete(1), spaces.Discrete(1))
        )

    def test_observation_space(self):
        assert self.env.observation_space == spaces.Tuple(
            (
                spaces.Box(-1e16, 1e16, (1,), dtype=np.float32),
                spaces.Box(-1e16, 1e16, (1,), dtype=np.float32),
            )
        )

    def test_step(self):
        observation, reward, terminated, truncated, info = self.env.step([0, 0])
        assert (observation == np.array([0.1])).all()

    def test_truncate(self):
        terminated = truncated = False
        while not (terminated or truncated):
            observation, reward, terminated, truncated, info = self.env.step([0, 0])
        assert truncated
        assert self.env.unwrapped.simulator.sim_time == 100.0



# ruff: noqa
import logging

fstr = "\x1b[30;3m%(asctime)s\x1b[0m %(shortname)-30s %(levelname)-10s <%(sim_time)-.2f> %(message)s"

colors = dict(
    GRAY=90,
    RED=91,
    GREEN=92,
    YELLOW=93,
    BLUE=94,
    MAGENTA=95,
    CYAN=96,
    WHITE=97,
    DARK_GRAY=30,
    DARK_RED=31,
    DARK_GREEN=32,
    DARK_YELLOW=33,
    DARK_BLUE=34,
    DARK_MAGENTA=35,
    DARK_CYAN=36,
    DARK_WHITE=37,
)


def style_string(
    string,
    no_format=False,
    style_spec=None,
    color=None,
    background_color=None,
    bold=False,
    emph=False,
    underline=False,
):
    if no_format:
        return string

    if style_spec is None:
        style_spec = []
    if color is not None:
        style_spec.append(colors[color.upper()])
    if background_color is not None:
        style_spec.append(colors[background_color.upper()] + 10)
    if bold:
        style_spec.append(1)
    if emph:
        style_spec.append(3)
    if underline:
        style_spec.append(4)
    return (
        "\x1b["
        + ";".join([str(style) for style in style_spec])
        + "m"
        + string
        + "\x1b[0m"
    )


level_color = {
    "DEBUG": None,
    "INFO": None,
    "WARNING": "YELLOW",
    "ERROR": "RED",
    "CRITICAL": "RED",
}
sat_color_cycle = [
    "DARK_CYAN",
    "GREEN",
    "DARK_BLUE",
    "MAGENTA",
    "CYAN",
    "DARK_GREEN",
    "BLUE",
    "DARK_MAGENTA",
]


class SimFormatter(logging.Formatter):
    def __init__(self, *args, color_output=True, **kwargs):
        super().__init__(
            *args,
            **kwargs,
        )
        self.color_output = color_output
        self.satellite_colors = {}

    def set_format(self, fmt, defaults=None, style="%"):
        self._style = logging._STYLES[style][0](fmt, defaults=defaults)
        self._fmt = self._style._fmt

    def format(self, record):
        no_format = not self.color_output
        fstr = ""
        fstr += style_string(
            "%(asctime)s ", color="GRAY", emph=True, no_format=no_format
        )

        sat_name = None
        sat_color = None
        try:
            if record.name.split(".")[1] == "sats":
                sat_name = record.name.split(".")[3]
                if sat_name not in self.satellite_colors:
                    self.satellite_colors[sat_name] = len(self.satellite_colors) % len(
                        sat_color_cycle
                    )
                sat_color = sat_color_cycle[self.satellite_colors[sat_name]]
        except IndexError:
            pass

        record.shortname = ".".join(record.name.split(".")[1:])
        fstr += style_string("%(shortname)-30s ", color=sat_color, no_format=no_format)
        fstr += style_string(
            "%(levelname)-10s ",
            color=level_color[record.levelname],
            bold=record.levelname == "CRITICAL",
            no_format=no_format,
        )
        if hasattr(record, "sim_time"):
            fstr += style_string(
                "<%(sim_time)-.2f> ", color="DARK_YELLOW", no_format=no_format
            )
        if sat_name is not None:
            fstr += style_string(
                f"{sat_name}: ",
                color=sat_color,
                no_format=no_format,
            )
        if record.msg.startswith("=== "):
            fstr += style_string(
                "%(message)s",
                bold=True,
                color="YELLOW",
                no_format=no_format,
            )
        else:
            fstr += style_string(
                "%(message)s",
                color=level_color[record.levelname],
                bold=record.levelname == "CRITICAL",
                no_format=no_format,
            )

        self.set_format(fstr)
        return super().format(record)


class ContextFilter(logging.Filter):
    def __init__(self, name: str = "", env=None, proc_id=None) -> None:
        super().__init__(name)
        self.env = env
        self.proc_id = proc_id

    def filter(self, record):
        try:
            record.sim_time = self.env.unwrapped.simulator.sim_time
        except AttributeError:
            pass
        return self.proc_id == record.process



"""Observations are found at ``bsk_rl.obs``.

Satellite observation types can be used to add information to the observation.
:class:`Observation` provides an interface for creating new observation types. To
configure the observation, set the ``observation_spec`` attribute of a
:class:`~bsk_rl.sats.Satellite` subclass. For example:

.. code-block:: python

    class MyObservationSatellite(Satellite):
        observation_spec = [
            SatProperties(
                dict(prop="r_BN_P", module="dynamics", norm=REQ_EARTH * 1e3),
                dict(prop="v_BN_P", module="dynamics", norm=7616.5),
            ),
            obs.OpportunityProperties(
                dict(prop="priority"),
                dict(prop="r_LP_P", norm=REQ_EARTH * 1e3),
                n_ahead_observe=16,
            ),
            obs.Time(),
        ]

The format of the observation can setting the ``obs_type`` attribute of the
:class:`~bsk_rl.sats.Satellite`. The default is ``np.ndarray``, but
it can also be set to a human-readable ``dict`` or a ``list``.

Some commonly used observations are provided:

* :class:`SatProperties` - Add arbitrary ``dynamics`` and ``fsw`` properties.
* :class:`RelativeProperties` - Add arbitrary properties relative to some other satellite.
* :class:`Time` - Add simulation time to the observation.
* :class:`OpportunityProperties` - Add information about upcoming targets or other ground access points to the observation.
* :class:`Eclipse` - Add a tuple of the next orbit start and end.
* :class:`ResourceRewardWeight` - Reports the weights of any randomized :class:`ResourceReward`.
"""

from bsk_rl.obs.observations import (
    Eclipse,
    Observation,
    OpportunityProperties,
    ResourceRewardWeight,
    SatProperties,
    Time,
)
from bsk_rl.obs.relative_observations import RelativeProperties

__doc_title__ = "Observations"
__all__ = [
    "Observation",
    "SatProperties",
    "RelativeProperties",
    "Time",
    "OpportunityProperties",
    "Eclipse",
    "ResourceRewardWeight",
]



from unittest.mock import MagicMock, patch

import numpy as np
import pytest
from pytest import approx

from bsk_rl.scene import CityTargets, UniformNadirScanning, UniformTargets
from bsk_rl.scene.targets import Target
from bsk_rl.utils.orbital import lla2ecef


class TestTarget:
    T1a = Target("Boulder", [1, 2, 3], 1.0)
    T1b = Target("Boulder", [1, 2, 3], 1.0)
    T2 = Target("Pleasanton", np.array([0, 0, 0]), 0.5)

    def test_id_unique(self):
        assert self.T1a.id != self.T1b.id

    def test_hash_unique(self):
        assert hash(self.T1a) != hash(self.T1b)

    def test_repr_(self):
        assert self.T2.__repr__() == "Target(Pleasanton)"

    def test_location_type(self):
        assert np.all(self.T1a.r_LP_P == np.array([1, 2, 3]))
        assert not isinstance(self.T1a.r_LP_P, list)
        assert np.all(self.T2.r_LP_P == np.array([0, 0, 0]))


class TestUniformTargets:
    def test_init(self):
        st = UniformTargets(1)
        assert st.priority_distribution is not None

    def test_reset_constant(self):
        st = UniformTargets(10)
        st.satellites = MagicMock()
        st.regenerate_targets = MagicMock()
        st.reset_pre_sim_init()
        assert st.n_targets == 10
        st.regenerate_targets.assert_called_once()

    @pytest.mark.repeat(10)
    def test_reset_variable(self):
        st = UniformTargets((8, 10))
        st.satellites = MagicMock()
        st.regenerate_targets = MagicMock()
        st.reset_pre_sim_init()
        assert 8 <= st.n_targets <= 10

    def test_regenerate_targets(self):
        st = UniformTargets(3, radius=1.0, priority_distribution=lambda: 1)
        st.satellites = MagicMock()
        st.n_targets = st._n_targets
        st.regenerate_targets()
        assert len(st.targets) == 3
        for target in st.targets:
            assert np.linalg.norm(target.r_LP_P) == approx(1.0)
            assert target.priority == 1

    def test_regenerate_targets_repeatable(self):
        np.random.seed(0)
        st1 = UniformTargets(3, radius=1.0)
        st1.satellites = MagicMock()
        st1.reset_pre_sim_init()
        np.random.seed(0)
        st2 = UniformTargets(3, radius=1.0)
        st2.satellites = MagicMock()
        st2.reset_pre_sim_init()
        for t1, t2 in zip(st1.targets, st2.targets):
            assert (t1.r_LP_P == t2.r_LP_P).all()


class TestCityTargets:
    @pytest.mark.parametrize(
        "lat,long,radius,expected",
        [
            (0, 0, 1, np.array([1, 0, 0])),
            (0, 0, 2, np.array([2, 0, 0])),
            (90, 0, 1, np.array([0, 0, 1])),
            (0, 180, 1, np.array([-1, 0, 0])),
        ],
    )
    def test_lla2ecef(self, lat, long, radius, expected):
        np.testing.assert_allclose(lla2ecef(lat, long, radius), expected, atol=1e-6)

    def mock_data(self, mock_read_csv, n_database=5):
        mock_read_csv.return_value = MagicMock(
            iloc=MagicMock(
                __getitem__=lambda self, i: {
                    "city": f"city{i}",
                    "lat": 0.0,
                    "lng": 0.0,
                    "iso2": "US",
                }
            ),
            __len__=lambda self: n_database,
        )

    @pytest.mark.parametrize(
        "n_targets",
        [0, 2, 5, 10],
    )
    @patch("pandas.read_csv")
    def test_regenerate_targets(self, mock_read_csv, n_targets):
        n_database = 5
        self.mock_data(mock_read_csv, n_database=n_database)
        ct = CityTargets(n_targets)
        ct.satellites = MagicMock()
        if n_targets > n_database:
            with pytest.raises(ValueError):
                ct.reset_pre_sim_init()
        else:
            ct.reset_pre_sim_init()
            assert len(ct.targets) == n_targets
            possible_names = [f"city{i}, US" for i in range(5)]
            for target in ct.targets:
                assert target.name in possible_names

    @pytest.mark.parametrize(
        "n_targets,n_select_from",
        [(0, 2), (1, 2), (2, 2), (3, "all")],
    )
    @patch("pandas.read_csv")
    def test_regenerate_targets_n_select_from(
        self, mock_read_csv, n_targets, n_select_from
    ):
        self.mock_data(mock_read_csv)
        ct = CityTargets(n_targets, n_select_from=n_select_from)
        ct.satellites = MagicMock()
        ct.reset_pre_sim_init()
        assert len(ct.targets) == n_targets
        if isinstance(n_select_from, int):
            possible_names = [f"city{i}, US" for i in range(n_select_from)]
            for target in ct.targets:
                assert target.name in possible_names

    @patch("bsk_rl.scene.targets.lla2ecef")
    @patch("pandas.read_csv")
    def test_regenerate_targets_offset(self, mock_read_csv, mock_lla2ecef):
        nominal = np.array([1.0, 0.0, 0.0])
        mock_lla2ecef.return_value = nominal
        self.mock_data(mock_read_csv, n_database=10)
        n_targets = 10
        ct = CityTargets(n_targets, location_offset=0.01, radius=1.0)
        ct.satellites = MagicMock()
        ct.reset_pre_sim_init()
        for target in ct.targets:
            assert np.linalg.norm(target.r_LP_P - nominal) <= 0.03
            assert np.linalg.norm(target.r_LP_P) == approx(1.0)


class TestUniformNadirScanning:
    def test_init(self):
        UniformNadirScanning()



import gymnasium as gym

from bsk_rl import act, data, obs, sats, scene, sim
from bsk_rl.data.composition import ComposedReward
from bsk_rl.utils.orbital import random_orbit

# For data models not tested in other tests

# NoData sufficiently checked in many cases

# UniqueImageData sufficiently checked in test_int_communication

# from ..test_int_full_environments


class FullFeaturedSatellite(sats.ImagingSatellite):
    observation_spec = [
        obs.SatProperties(dict(prop="r_BN_P", module="dynamics", norm=6e6)),
        obs.Time(),
    ]
    action_spec = [act.Image(n_ahead_image=10)]


def test_multi_rewarder():
    env = gym.make(
        "GeneralSatelliteTasking-v1",
        satellites=[
            FullFeaturedSatellite(
                "Sentinel-2A",
                sat_args=FullFeaturedSatellite.default_sat_args(
                    oe=random_orbit,
                    imageAttErrorRequirement=0.01,
                    imageRateErrorRequirement=0.01,
                ),
            ),
            FullFeaturedSatellite(
                "Sentinel-2B",
                sat_args=FullFeaturedSatellite.default_sat_args(
                    oe=random_orbit,
                    imageAttErrorRequirement=0.01,
                    imageRateErrorRequirement=0.01,
                ),
            ),
        ],
        scenario=scene.UniformTargets(n_targets=1000),
        rewarder=(data.UniqueImageReward(), data.UniqueImageReward()),
        sim_rate=0.5,
        max_step_duration=1e9,
        time_limit=5700.0,
        disable_env_checker=True,
    )

    assert isinstance(env.unwrapped.rewarder, ComposedReward)

    env.reset()
    for _ in range(10):
        env.step(env.action_space.sample())


class DriftSat(sats.Satellite):
    observation_spec = [obs.Time()]
    action_spec = [act.Drift(duration=100.0)]
    dyn_type = sim.dyn.BasicDynamicsModel
    fsw_type = sim.fsw.BasicFSWModel


def test_time_penalty():
    env = gym.make(
        "SatelliteTasking-v1",
        satellite=DriftSat(
            "DriftSat",
        ),
        rewarder=data.ResourceReward(
            reward_weight=-0.1,
            resource_fn=lambda sat: sat.simulator.sim_time,
        ),
        disable_env_checker=True,
    )

    env.reset()
    _, reward, _, _, _ = env.step(0)
    assert reward == -10.0



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html


import datetime
import importlib

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
import os
import re
import shutil
import sys
from importlib import metadata
from pathlib import Path

from nbdime.diffing.notebooks import diff_notebooks
from nbdime.utils import read_notebook

# sys.path.insert(0, os.path.abspath(os.path.join("..", "..", "src")))
now = datetime.datetime.now()

release = metadata.version("bsk_rl")
project = "BSK-RL v" + release
copyright = str(now.year) + ", Autonomous Vehicle Systems (AVS) Laboratory"
author = "Mark Stephenson"
version = "Version " + release

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.doctest",
    "sphinx.ext.todo",
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx_rtd_theme",
    "nbsphinx",
    "sphinx.ext.mathjax",
    "sphinxcontrib.youtube",
]

templates_path = ["_templates"]
exclude_patterns = []
source_suffix = ".rst"
master_doc = "index"
autoclass_content = "init"
autodoc_member_order = "bysource"
autodoc_default_options = {
    "undoc-members": None,
}
autodoc_typehints = "both"
# nbsphinx_execute = "never"
nbsphinx_allow_errors = False


nitpicky = True
nitpick_ignore = [
    ("py:obj", "bsk_rl.gym.SatAct"),
    ("py:obj", "bsk_rl.gym.SatObs"),
    ("py:class", "bsk_rl.gym.SatAct"),
    ("py:class", "bsk_rl.gym.SatObs"),
]
nitpick_ignore_regex = [
    ("py:class", r"^(?!bsk_rl).*"),
    ("py:obj", r"^(?!bsk_rl).*"),
]


# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_theme_options = {
    "style_nav_header_background": "#CFB87C",
    "navigation_depth": -1,
}
html_static_path = ["_static"]
html_css_files = ["custom.css"]
html_logo = "./_images/static/bsk_rl-logo.png"

add_module_names = False


def skip(app, what, name, obj, would_skip, options):
    return would_skip


def setup(app):
    app.connect("autodoc-skip-member", skip)


def has_all(qual_name):
    try:
        all = importlib.import_module(qual_name).__all__
        if len(all) == 0:
            return False
    except AttributeError:
        return False

    return True


class PackageCrawler:
    def __init__(
        self, base_source_dir, base_doc_dir, filter_all=True, nb_cache_dir=None
    ):
        self.base_source_dir = base_source_dir
        self.base_doc_dir = base_doc_dir
        self.filter_all = filter_all
        self.nb_cache_dir = nb_cache_dir

    def grab_files(self, dir_path):
        dirs_in_dir = [x for x in dir_path.iterdir() if x.is_dir()]
        dir_filters = [
            r".*__pycache__.*",
            r".*\.ruff_cache.*",
            r".*\.egg-info",
            r".*\/simplemaps_worldcities",
            r".*\/_.*",
        ]
        dirs_in_dir = list(
            filter(
                lambda dir: not any(
                    re.match(filter, str(dir)) for filter in dir_filters
                ),
                dirs_in_dir,
            )
        )

        files_in_dir = dir_path.glob("*.py")
        file_filters = [
            r".*__init__\.py",
            r"(.*\/|)_[a-zA-Z0-9_]*\.py",
        ]
        files_in_dir = list(
            filter(
                lambda file: not any(
                    re.match(filter, str(file)) for filter in file_filters
                ),
                files_in_dir,
            )
        )

        if self.filter_all:
            files_in_dir = list(
                filter(
                    lambda file: has_all(
                        str(file.relative_to(self.base_source_dir.parent))
                        .replace("/", ".")
                        .replace(".py", "")
                    ),
                    files_in_dir,
                )
            )

        notebooks_in_dir = dir_path.glob("*.ipynb")

        return (
            sorted(list(files_in_dir)),
            sorted(list(dirs_in_dir)),
            sorted(list(notebooks_in_dir)),
        )

    def generate_index(self, index_path, file_paths, dir_paths, source_dir):
        # Make header
        lines = ""
        qual_name = str(source_dir.relative_to(self.base_source_dir.parent)).replace(
            "/", "."
        )
        lines += ".. _" + qual_name.replace(" ", "_") + ":\n"
        lines += f".. currentmodule:: {qual_name}\n\n"

        # if a _default.rst file exists in a folder, then use it to generate the index.rst file
        try:
            docFileName = source_dir / "_default.rst"
            with open(docFileName, "r") as docFile:
                docContents = docFile.read()
            lines += docContents + "\n\n"
        except FileNotFoundError:  # Auto-generate the index.rst file
            # add page tag

            # pull in folder _doc.rst file if it exists
            docContents = ""
            try:
                docFileName = source_dir / "_doc.rst"
                if os.path.isfile(docFileName):
                    with open(docFileName, "r") as docFile:
                        docContents = docFile.read()
            except FileNotFoundError:
                pass

            # Title the page
            if "\n===" not in docContents:
                # Make title
                doc_title = qual_name
                try:
                    doc_title = importlib.import_module(qual_name).__doc_title__
                except AttributeError:
                    pass
                lines += doc_title + "\n" + "=" * len(doc_title) + "\n\n"

            lines += docContents + "\n\n"

            # Also check for docs in the __init__.py file
            # lines += "\n\nReference\n----------\n\n"
            lines += (
                """.. automodule:: """
                + qual_name
                + """\n   :members:\n   :show-inheritance:\n\n"""
            )

            if len(file_paths) > 0 or len(dir_paths) > 0:
                # lines += "\n\nModules\n----------\n\n"
                pass

            # Add a linking point to all local files
            lines += "\n\n.. toctree::\n   :maxdepth: 1\n   :hidden:\n\n"
            added_names = []
            for file_path in sorted(file_paths):
                file_name = os.path.basename(os.path.normpath(file_path))
                file_name = file_name[: file_name.rfind(".")]

                if file_name not in added_names:
                    lines += "   " + file_name + "\n"
                    added_names.append(file_name)
            lines += "\n"

            # Add a linking point to all local directories
            lines += ".. toctree::\n   :maxdepth: 1\n   :hidden:\n\n"

            for dir_path in sorted(dir_paths):
                dirName = os.path.basename(os.path.normpath(dir_path))
                lines += "   " + dirName + "/index\n"

        with open(os.path.join(index_path, "index.rst"), "w") as f:
            f.write(lines)

    def generate_autodoc(self, doc_path, source_file):
        # Make header
        short_name = source_file.name.replace(".py", "")
        lines = ""
        qual_name = (
            str(source_file.relative_to(self.base_source_dir.parent))
            .replace("/", ".")
            .replace(".py", "")
        )
        lines += ".. _" + qual_name.replace(" ", "_") + ":\n"
        lines += f".. currentmodule:: {qual_name}\n\n"

        # Make title
        doc_title = short_name
        try:
            doc_title = importlib.import_module(qual_name).__doc_title__
        except (AttributeError, ModuleNotFoundError):
            pass
        lines += doc_title + "\n" + "=" * len(doc_title) + "\n\n"

        # Generate the autodoc file
        # lines += f"``{qual_name}``\n\n"
        lines += ".. toctree::\n   :maxdepth: 1\n\n"
        lines += (
            """.. automodule:: """
            + qual_name
            + """\n   :members:\n   :show-inheritance:\n\n"""
        )

        # Write to file
        with open(doc_path / f"{short_name}.rst", "w") as f:
            f.write(lines)

    def run(self, source_dir=None):
        if source_dir is None:
            source_dir = self.base_source_dir

        file_paths, dir_paths, nb_paths = self.grab_files(source_dir)
        index_path = source_dir.relative_to(self.base_source_dir)

        # Populate the index.rst file of the local directory
        os.makedirs(self.base_doc_dir / index_path, exist_ok=True)
        self.generate_index(
            self.base_doc_dir / index_path, file_paths + nb_paths, dir_paths, source_dir
        )

        # Generate the correct auto-doc function for python modules
        for file in file_paths:
            self.generate_autodoc(
                self.base_doc_dir / index_path,
                file,
            )

        for notebook in nb_paths:
            nb_cache = (
                self.nb_cache_dir / self.base_doc_dir / index_path / notebook.name
            )
            if (
                self.nb_cache_dir is not None
                and nb_cache.is_file()
                and (
                    "'source'"
                    not in diff_notebooks(
                        read_notebook(notebook.resolve(), on_null="empty"),
                        read_notebook(nb_cache.resolve(), on_null="empty"),
                    ).__repr__()
                )
                and (
                    "'source'"
                    not in diff_notebooks(
                        read_notebook(nb_cache.resolve(), on_null="empty"),
                        read_notebook(notebook.resolve(), on_null="empty"),
                    ).__repr__()
                )
            ):
                shutil.copy(
                    nb_cache,
                    self.base_doc_dir / index_path / notebook.name,
                )
            else:
                shutil.copy(notebook, self.base_doc_dir / index_path / notebook.name)

        # Recursively go through all directories in source, documenting what is available.
        for dir_path in sorted(dir_paths):
            self.run(
                source_dir=dir_path,
            )

        return


sys.path.append(os.path.abspath("../.."))
nb_cache_dir = Path("../build/doctrees/nbsphinx")
PackageCrawler(
    Path("../../src/bsk_rl/"), Path("./api_reference/"), nb_cache_dir=nb_cache_dir
).run()
PackageCrawler(
    Path("../../examples"),
    Path("./examples/"),
    filter_all=False,
    nb_cache_dir=nb_cache_dir,
).run()



import gymnasium as gym
from gymnasium.envs.registration import register

from bsk_rl.gym import (
    NO_ACTION,
    ConstellationTasking,
    GeneralSatelliteTasking,
    SatelliteTasking,
)

__all__ = [
    "GeneralSatelliteTasking",
    "SatelliteTasking",
    "ConstellationTasking",
]

register(
    id="GeneralSatelliteTasking-v1",
    entry_point="bsk_rl.gym:GeneralSatelliteTasking",
)

register(
    id="SatelliteTasking-v1",
    entry_point="bsk_rl.gym:SatelliteTasking",
)

register(
    id="ConstellationTasking-v1",
    entry_point="bsk_rl.gym:ConstellationTasking",
)



"""Data system for recording unique images of targets."""

import logging
from typing import TYPE_CHECKING, Callable, Optional

import numpy as np

from bsk_rl.data.base import Data, DataStore, GlobalReward
from bsk_rl.utils import vizard

if TYPE_CHECKING:
    from bsk_rl.sats import Satellite
    from bsk_rl.scene.targets import Target

logger = logging.getLogger(__name__)


class UniqueImageData(Data):
    """Data for unique images of targets."""

    def __init__(
        self,
        imaged: Optional[set["Target"]] = None,
        duplicates: int = 0,
        known: Optional[set["Target"]] = None,
    ) -> None:
        """Construct unit of data to record unique images.

        Keeps track of ``imaged`` targets, a count of ``duplicates`` (i.e. images that
        were not rewarded due to the target already having been imaged), and all
        ``known`` targets in the environment.

        Args:
            imaged: Set of targets that are known to be imaged.
            duplicates: Count of target imaging duplication.
            known: Set of targets that are known to exist (imaged and unimaged).
        """
        if imaged is None:
            imaged = set()
        self.imaged = set(imaged)
        self.duplicates = duplicates + len(imaged) - len(self.imaged)
        if known is None:
            known = set()
        self.known = set(known)

    def __add__(self, other: "UniqueImageData") -> "UniqueImageData":
        """Combine two units of data.

        Args:
            other: Another unit of data to combine with this one.

        Returns:
            Combined unit of data.
        """
        imaged = self.imaged | other.imaged
        duplicates = (
            self.duplicates
            + other.duplicates
            + len(self.imaged)
            + len(other.imaged)
            - len(imaged)
        )
        known = self.known | other.known
        return self.__class__(imaged=imaged, duplicates=duplicates, known=known)


class UniqueImageStore(DataStore):
    """DataStore for unique images of targets."""

    data_type = UniqueImageData

    def __init__(self, *args, **kwargs) -> None:
        """DataStore for unique images.

        Detects new images by watching for an increase in data in each target's corresponding
        buffer.
        """
        super().__init__(*args, **kwargs)

    def get_log_state(self) -> float:
        """Log the instantaneous storage unit state at the end of each step.

        Returns:
            float: storedData from satellite storage unit
        """
        msg = self.satellite.dynamics.storageUnit.storageUnitDataOutMsg.read()
        return msg.storedData[0]

    def compare_log_states(
        self, old_state: np.ndarray, new_state: np.ndarray
    ) -> UniqueImageData:
        """Check for an increase in logged data to identify new images.

        Args:
            old_state: Older storedData from satellite storage unit.
            new_state: Newer storedData from satellite storage unit.

        Returns:
            list: Targets imaged at new_state that were unimaged at old_state.
        """
        data_increase = new_state - old_state
        if data_increase <= 0:
            return UniqueImageData()
        else:
            assert self.satellite.latest_target is not None
            self.update_target_colors([self.satellite.latest_target])
            return UniqueImageData(imaged={self.satellite.latest_target})

    @vizard.visualize
    def update_target_colors(self, targets, vizInstance=None, vizSupport=None):
        """Update target colors in Vizard."""
        for target in targets:
            vizSupport.changeLocation(
                vizInstance,
                target.name,
                color=vizSupport.toRGBA255(self.satellite.vizard_color),
            )


class UniqueImageReward(GlobalReward):
    """GlobalReward for rewarding unique images."""

    data_store_type = UniqueImageStore

    def __init__(
        self,
        reward_fn: Callable = lambda p: p,
    ) -> None:
        """GlobalReward for rewarding unique images.

        This data system should be used with the :class:`~bsk_rl.sats.ImagingSatellite` and
        a scenario that generates targets, such as :class:`~bsk_rl.scene.UniformTargets` or
        :class:`~bsk_rl.scene.CityTargets`.

        The satellites all start with complete knowledge of the targets in the scenario.
        Each target can only give one satellite a reward once; if any satellite has imaged
        a target, reward will never again be given for that target. The satellites filter
        known imaged targets from consideration for imaging to prevent duplicates.
        Communication can transmit information about what targets have been imaged in order
        to prevent reimaging.


        Args:
            scenario: GlobalReward.scenario
            reward_fn: Reward as function of priority.
        """
        super().__init__()
        self.reward_fn = reward_fn

    def initial_data(self, satellite: "Satellite") -> "UniqueImageData":
        """Furnish data to the scenario.

        Currently, it is assumed that all targets are known a priori, so the initial data
        given to the data store is the list of all targets.
        """
        return self.data_type(known=self.scenario.targets)

    def create_data_store(self, satellite: "Satellite") -> None:
        """Override the access filter in addition to creating the data store."""
        super().create_data_store(satellite)

        def unique_target_filter(opportunity):
            if opportunity["type"] == "target":
                return opportunity["object"] not in satellite.data_store.data.imaged
            return True

        satellite.add_access_filter(unique_target_filter)

    def calculate_reward(
        self, new_data_dict: dict[str, UniqueImageData]
    ) -> dict[str, float]:
        """Reward each new unique image once.

        Reward is evaluated based on ``self.reward_fn(target.priority)``.

        Args:
            new_data_dict: Record of new images for each satellite

        Returns:
            reward: Cumulative reward across satellites for one step
        """
        reward = {}
        imaged_counts = {}
        for new_data in new_data_dict.values():
            for target in new_data.imaged:
                if target not in imaged_counts:
                    imaged_counts[target] = 0
                imaged_counts[target] += 1

        for sat_id, new_data in new_data_dict.items():
            reward[sat_id] = 0.0
            for target in new_data.imaged:
                if target not in self.data.imaged:
                    reward[sat_id] += (
                        self.reward_fn(target.priority) / imaged_counts[target]
                    )
        return reward


__doc_title__ = "Unique Images"
__all__ = ["UniqueImageReward", "UniqueImageStore", "UniqueImageData"]



"""Satellite action types can be used to add actions to an agent."""

from abc import ABC, abstractmethod
from copy import deepcopy
from typing import TYPE_CHECKING, Any

from bsk_rl.utils.functional import AbstractClassProperty, Resetable

if TYPE_CHECKING:  # pragma: no cover
    from gymnasium import spaces

    from bsk_rl.sats import Satellite
    from bsk_rl.sim import Simulator


def select_action_builder(satellite: "Satellite") -> "ActionBuilder":
    """Identify the proper action builder based on a satellite's action spec.

    Args:
        satellite: Satellite to build actions for.

    Returns:
        action builder of the appropriate type
    """
    builder_types = [spec.builder_type for spec in satellite.action_spec]
    if all([builder_type == builder_types[0] for builder_type in builder_types]):
        return builder_types[0](satellite)
    else:
        raise NotImplementedError("Heterogenous action builders not supported.")


class ActionBuilder(ABC, Resetable):

    def __init__(self, satellite: "Satellite") -> None:
        """Base class for all action builders.

        Args:
            satellite: Satellite to build actions for.
        """
        self.satellite = satellite
        self.simulator: "Simulator"
        self.action_spec = deepcopy(self.satellite.action_spec)
        for act in self.action_spec:
            act.link_satellite(self.satellite)

    def reset_overwrite_previous(self) -> None:
        """Perform any once-per-episode setup."""
        for act in self.action_spec:
            act.reset_overwrite_previous()

    def reset_pre_sim_init(self) -> None:
        """Perform any once-per-episode setup."""
        for act in self.action_spec:
            act.reset_pre_sim_init()

    def reset_during_sim_init(self) -> None:
        """Perform any once-per-episode setup."""
        for act in self.action_spec:
            act.reset_during_sim_init()

    def reset_post_sim_init(self) -> None:
        """Perform any once-per-episode setup."""
        self.simulator = self.satellite.simulator  # already a proxy
        for act in self.action_spec:
            act.link_simulator(self.simulator)  # already a proxy
            act.reset_post_sim_init()

    @property
    @abstractmethod
    def action_space(self) -> "spaces.Space":
        """Return the action space."""
        pass

    @property
    @abstractmethod
    def action_description(self) -> Any:
        """Return a description of the action space."""
        pass

    @abstractmethod
    def set_action(self, action: Any) -> None:
        """Set the action to be taken."""
        pass


class Action(ABC, Resetable):
    builder_type: type[ActionBuilder] = AbstractClassProperty()  #: :meta private:

    def __init__(self, name: str = "act") -> None:
        """Base class for all actions.

        Args:
            name: Name of the action.
        """
        self.name = name
        self.satellite: "Satellite"
        self.simulator: "Simulator"

    def link_satellite(self, satellite: "Satellite") -> None:
        """Link the action to a satellite.

        Args:
            satellite: Satellite to link to

        :meta private:
        """
        self.satellite = satellite  # already a proxy

    def link_simulator(self, simulator: "Simulator") -> None:
        """Link the action to a simulator.

        Args:
            simulator: Simulator to link to

        :meta private:
        """
        self.simulator = simulator  # already a proxy

    @abstractmethod
    def set_action(self, action: Any) -> None:  # pragma: no cover
        """Execute code to perform an action."""
        pass


__doc_title__ = "Backend"
__all__ = ["ActionBuilder"]



from unittest.mock import MagicMock, call, patch

import numpy as np
import pytest

from bsk_rl.sim.world import BasicWorldModel, GroundStationWorldModel, WorldModel

module = "bsk_rl.sim.world."


class TestWorldModel:
    @patch(module + "collect_default_args")
    def test_default_world_args(self, mock_collect):
        mock_collect.return_value = {"a": 1, "b": 2, "c": 3}
        assert WorldModel.default_world_args() == {"a": 1, "b": 2, "c": 3}

    @pytest.mark.parametrize(
        "overwrite,error", [({"c": 4}, False), ({"not_c": 4}, True)]
    )
    @patch(module + "collect_default_args")
    def test_default_sat_args_overwrote(self, mock_collect, overwrite, error):
        mock_collect.return_value = {"a": 1, "b": 2, "c": 3}
        if not error:
            assert WorldModel.default_world_args(**overwrite) == {
                "a": 1,
                "b": 2,
                "c": 4,
            }
        else:
            with pytest.raises(KeyError):
                WorldModel.default_world_args(**overwrite)

    @patch.multiple(WorldModel, __abstractmethods__=set())
    @patch(module + "WorldModel._setup_world_objects")
    def test_init(self, mock_obj_init):
        mock_sim = MagicMock()
        kwargs = dict(a=1, b=2)
        world = WorldModel(mock_sim, 1.0, **kwargs)
        mock_sim.CreateNewProcess.assert_called_once()
        mock_sim.CreateNewTask.assert_called_once()
        mock_obj_init.assert_called_once_with(**kwargs)
        assert world.simulator is not mock_sim
        assert world.simulator == mock_sim


class TestBasicWorldModel:
    basicworld = module + "BasicWorldModel."

    @patch(basicworld + "__init__", MagicMock(return_value=None))
    def test_PN(self):
        world = BasicWorldModel(MagicMock(), 1.0)
        world.gravFactory = MagicMock()
        world.body_index = 0
        msg = world.gravFactory.spiceObject.planetStateOutMsgs.__getitem__
        msg.return_value.read.return_value.J20002Pfix = [1, 0, 0, 0, 1, 0, 0, 0, 1]
        assert (world.PN == np.identity(3)).all()

    @patch(basicworld + "__init__", MagicMock(return_value=None))
    def test_omega_PN_N(self):
        world = BasicWorldModel(MagicMock(), 1.0)
        world.gravFactory = MagicMock()
        world.body_index = 0
        msg = world.gravFactory.spiceObject.planetStateOutMsgs.__getitem__
        msg.return_value.read.return_value.J20002Pfix_dot = [1, 0, 0, 0, 1, 0, 0, 0, 1]
        msg.return_value.read.return_value.J20002Pfix = [0, -1, 1, 1, 0, -1, -1, 1, 0]
        assert (world.omega_PN_N == np.array([1, 1, 1])).all()

    @patch(basicworld + "setup_gravity_bodies")
    @patch(basicworld + "setup_ephem_object")
    @patch(basicworld + "setup_atmosphere_density_model")
    @patch(basicworld + "setup_eclipse_object")
    def test_setup_and_delete(self, grav_set, epoch_set, atmos_set, eclipse_set):
        world = BasicWorldModel(MagicMock(), 1.0)
        for setter in (grav_set, epoch_set, atmos_set, eclipse_set):
            setter.assert_called_once()
        unload_function = MagicMock()
        world.gravFactory = MagicMock(unloadSpiceKernels=unload_function)
        del world
        unload_function.assert_called_once()

    @patch(basicworld + "_setup_world_objects", MagicMock())
    @patch(module + "simIncludeGravBody", MagicMock())
    def testsetup_gravity_bodies(self):
        # Smoke test
        world = BasicWorldModel(MagicMock(), 1.0)
        world.simulator = MagicMock()
        world.setup_gravity_bodies(utc_init="time")
        world.simulator.AddModelToTask.assert_called_once()

    @patch(basicworld + "_setup_world_objects", MagicMock())
    @patch(module + "ephemerisConverter", MagicMock())
    def testsetup_epoch_object(self):
        # Smoke test
        world = BasicWorldModel(MagicMock(), 1.0)
        world.simulator = MagicMock()
        world.gravFactory = MagicMock()
        world.sun_index = 0
        world.body_index = 1
        world.setup_ephem_object()
        world.simulator.AddModelToTask.assert_called_once()

    @patch(basicworld + "_setup_world_objects", MagicMock())
    @patch(module + "exponentialAtmosphere", MagicMock())
    def testsetup_atmosphere_density_model(self):
        # Smoke test
        world = BasicWorldModel(MagicMock(), 1.0)
        world.simulator = MagicMock()
        world.gravFactory = MagicMock()
        world.body_index = 1

        world.setup_atmosphere_density_model(
            planetRadius=1.0,
            baseDensity=1.0,
            scaleHeight=1.0,
        )
        world.simulator.AddModelToTask.assert_called_once()

    @patch(basicworld + "_setup_world_objects", MagicMock())
    @patch(module + "eclipse", MagicMock())
    def testsetup_eclipse_object(self):
        # Smoke test
        world = BasicWorldModel(MagicMock(), 1.0)
        world.simulator = MagicMock()
        world.gravFactory = MagicMock()
        world.sun_index = 0
        world.body_index = 1
        world.setup_eclipse_object()
        world.simulator.AddModelToTask.assert_called_once()


class TestGroundStationWorldModel:
    groundworld = module + "GroundStationWorldModel."

    @patch(groundworld + "setup_ground_locations")
    @patch(module + "BasicWorldModel._setup_world_objects", MagicMock())
    def test_setup_world_objects(self, ground_set):
        GroundStationWorldModel(MagicMock(), 1.0)
        ground_set.assert_called_once()

    @patch(groundworld + "_setup_world_objects", MagicMock())
    @patch(groundworld + "_create_ground_station")
    def testsetup_ground_locations(self, mock_gs_create):
        world = GroundStationWorldModel(MagicMock(), 1.0)
        world.setup_ground_locations([dict(a=1), dict(b=2)], 1000.0, 1.0, 1000.0)
        mock_gs_create.assert_has_calls(
            [call(a=1, priority=1399), call(b=2, priority=1398)]
        )

    @patch(groundworld + "_setup_world_objects", MagicMock())
    @patch(module + "groundLocation", MagicMock())
    def test_create_ground_station(self):
        world = GroundStationWorldModel(MagicMock(), 1.0)
        world.simulator = MagicMock()
        world.gravFactory = MagicMock()
        world.groundStations = []
        world.groundLocationPlanetRadius = 10.0
        world.gsMinimumElevation = 1.0
        world.gsMaximumRange = 1000.0
        world.body_index = 1
        world._create_ground_station(0.0, 0.0)
        world.simulator.AddModelToTask.assert_called_once()



"""Data collection and reward calculation is given in ``bsk_rl.data``.

Reward System Components
------------------------

The reward system has three main components: :class:`GlobalReward`, :class:`~bsk_rl.data.base.DataStore`,
and :class:`~bsk_rl.data.base.Data`.

:class:`GlobalReward` acts as a global critic for the environment, rewarding each
agent's performance. Has full knowledge of the environment and can provide rewards
based on the global state of the environment, even if the agent does not have access
to that information; for example, you may not want to reward an agent for imaging a
target that has already been imaged by another agent, even if the agent does not know
that the target has been imaged. Reward is generally calculated by processing the
dictionary of new :class:`~bsk_rl.data.base.Data` per-satellite generated at each step
with the :class:`GlobalReward.calculate_reward` method.

The :class:`~bsk_rl.data.base.DataStore` handles each satellite's local knowledge of the
scenario and the data it generates. The data store gains data in three ways:

1. On environment reset, the :class:`~bsk_rl.data.GlobalReward` calls
   :class:`~bsk_rl.data.GlobalReward.initial_data` to provide the initial knowledge of the
   scenario for each satellite. This may be empty or may contain some a priori knowledge,
   such as a list of targets that are desired to be imaged.
2. At the end of each step, the result of :class:`~bsk_rl.data.base.DataStore.get_log_state`
   is compared to the previous step's result via :class:`~bsk_rl.data.base.DataStore.compare_log_states`.
   A unit of :class:`~bsk_rl.data.base.Data` is returned. For example, the log state may
   be the level of each target's buffer partition in the storage unit, so a change in
   a certain buffer level leads to a unit of data that indicates the corresponding target
   has been imaged.
3. At the end of each step, satellites communicate based on the :ref:`bsk_rl.comm`
   system being used. Satellites merge the contents of their data stores with any other
   satellite's data store that they have communicated with.

Finally, :class:`~bsk_rl.data.base.Data` can represent data generated by the satellite
towards some goal (e.g. images of targets, time spend in a desireable mode, etc.) as well
as information about the environment that is useful toward completing its mission (e.g.
desired targets to image, what targets have already been imaged, etc.).

Implementing Data & Reward Types
================================

See :ref:`bsk_rl.data.base` for full documentation of the reward system components to
when implementing custom data and reward types.

Reward System Types
-------------------

A variety of reward systems are available for use in the environment. The following table
provides a summary of the available reward systems:

+-----------------------------+-------------------------------------------------------------------------+---------------------------------------------------------------------+
| **Type**                    | **Purpose**                                                             | **Compatibility**                                                   |
+-----------------------------+-------------------------------------------------------------------------+---------------------------------------------------------------------+
| :class:`NoReward`           | Returns zero reward for every agent at every step.                      |                                                                     |
+-----------------------------+-------------------------------------------------------------------------+---------------------------------------------------------------------+
| :class:`UniqueImageReward`  | Returns reward corresponding to target priority the                     | Should be used with :class:`~bsk_rl.sats.ImagingSatellite` and a    |
|                             | first time a target is imaged by any agent. Causes                      | :class:`~bsk_rl.scene.targets.Target`-based scenario.               |
|                             | satellites to filter targets that are known to have                     |                                                                     |
|                             | been imaged already.                                                    |                                                                     |
+-----------------------------+-------------------------------------------------------------------------+---------------------------------------------------------------------+
| :class:`ScanningTimeReward` | Returns reward based on time spend in the nadir-pointing scanning mode. | Should be used with the :class:`~bsk_rl.scene.UniformNadirScanning` |
|                             |                                                                         | scenario.                                                           |
+-----------------------------+-------------------------------------------------------------------------+---------------------------------------------------------------------+
| :class:`ResourceReward`     | Returns reward based on the change in an arbitrary resource level       |                                                                     |
|                             | (e.g. fuel, power, time, etc.).                                         |                                                                     |
+-----------------------------+-------------------------------------------------------------------------+---------------------------------------------------------------------+
| :class:`RSOInspectionReward`| Returns reward based on inspection percentage of an RSO                 | Works with :class:`RSOPoints` scenario.                             |
+-----------------------------+-------------------------------------------------------------------------+---------------------------------------------------------------------+



To select a reward system to use, pass an instance of :class:`GlobalReward` to the ``data``
field of the environment constructor:

.. code-block:: python

    env = ConstellationTasking(
        ...,
        data=ScanningTimeReward(),
        ...
    )

Multiple reward systems can be added to the environment by instead passing an iterable of
reward systems to the ``data`` field of the environment constructor:

.. code-block:: python

    env = ConstellationTasking(
        ...,
        data=(ScanningTimeReward(), SomeOtherReward()),
        ...
    )

On the backend, this creates a :class:`~bsk_rl.data.composition.ComposedDataStore` that
handles the combination of multiple reward systems.
"""

from bsk_rl.data.base import GlobalReward
from bsk_rl.data.nadir_data import ScanningTimeReward
from bsk_rl.data.no_data import NoReward
from bsk_rl.data.resource_data import ResourceReward
from bsk_rl.data.rso_inspection import RSOInspectionReward
from bsk_rl.data.unique_image_data import UniqueImageReward

__doc_title__ = "Data & Reward"
__all__ = [
    "GlobalReward",
    "NoReward",
    "UniqueImageReward",
    "ScanningTimeReward",
    "ResourceReward",
    "RSOInspectionReward",
]



"""Data composition classes."""

import logging
from typing import TYPE_CHECKING, Optional

from bsk_rl.data.base import Data, DataStore, GlobalReward
from bsk_rl.sats import Satellite
from bsk_rl.scene.scenario import Scenario

if TYPE_CHECKING:
    from bsk_rl.sats import Satellite

logger = logging.getLogger(__name__)


class ComposedData(Data):
    """Data for composed data types."""

    def __init__(self, *data: Data) -> None:
        """Data for composed data types.

        Args:
            data: Data types to compose.
        """
        self.data = data

    def __add__(self, other: "ComposedData") -> "ComposedData":
        """Combine two units of composed data.

        Args:
            other: Another unit of composed data to combine with this one.

        Returns:
            Combined unit of composed data.
        """
        if len(self.data) == 0 and len(other.data) == 0:
            data = []
        elif len(self.data) == 0:
            data = [type(d)() + d for d in other.data]
        elif len(other.data) == 0:
            data = [d + type(d)() for d in self.data]
        elif len(self.data) == len(other.data):
            data = [d1 + d2 for d1, d2 in zip(self.data, other.data)]
        else:
            raise ValueError(
                "ComposedData units must have the same number of data types."
            )
        return ComposedData(*data)

    def __getattr__(self, name: str):
        """Search for an attribute in the datas."""
        for data in self.data:
            if hasattr(data, name):
                return getattr(data, name)
        raise AttributeError(f"No Data in ComposedData has attribute '{name}'")

    def __repr__(self) -> str:
        """String representation of the ComposedData."""
        return f"ComposedData({', '.join(repr(d) for d in self.data)})"


class ComposedDataStore(DataStore):
    data_type = ComposedData

    def pass_data(self) -> None:
        """Pass data to the sub-DataStores.

        :meta private:
        """
        for ds, data in zip(self.data_stores, self.data.data):
            ds.data = data

    def __init__(
        self,
        satellite: "Satellite",
        *data_store_types: type[DataStore],
        initial_data: Optional[ComposedData] = None,
        data_store_kwargs: Optional[list] = None,
    ):
        """DataStore for composed data types.

        Args:
            satellite: Satellite which data is being stored for.
            data_store_types: DataStore types to compose.
            initial_data: Initial data to start the store with. Usually comes from
                :class:`~bsk_rl.data.GlobalReward.initial_data`.
            data_store_kwargs: List of data_store kwargs matching data_store_types.
        """
        self.data: ComposedData
        super().__init__(satellite, initial_data)
        if data_store_kwargs is None:
            data_store_kwargs = [{} for _ in data_store_types]

        if len(data_store_types) != len(data_store_kwargs):
            raise ValueError(
                "data_store_types and data_store_kwargs must have the same length."
            )

        self.data_stores = tuple(
            ds(satellite, **kwargs)
            for ds, kwargs in zip(data_store_types, data_store_kwargs)
        )
        self.pass_data()

    def __getattr__(self, name: str):
        """Search for an attribute in the data_stores."""
        for data_store in self.data_stores:
            if hasattr(data_store, name):
                return getattr(data_store, name)
        raise AttributeError(
            f"No DataStore in ComposedDataStore has attribute '{name}'"
        )

    def get_log_state(self) -> list:
        """Pull information used in determining current data contribution."""
        log_states = [ds.get_log_state() for ds in self.data_stores]
        return log_states

    def compare_log_states(self, prev_state: list, new_state: list) -> Data:
        """Generate a unit of composed data based on previous step and current step logs."""
        data = [
            ds.compare_log_states(prev, new)
            for ds, prev, new in zip(self.data_stores, prev_state, new_state)
        ]
        return ComposedData(*data)

    def update_from_logs(self) -> Data:
        """Update the data store based on collected information."""
        new_data = super().update_from_logs()
        self.pass_data()
        return new_data

    def update_with_communicated_data(self) -> None:
        """Update the data store based on collected information from other satellites."""
        super().update_with_communicated_data()
        self.pass_data()


class ComposedReward(GlobalReward):
    data_store_type = ComposedDataStore

    def pass_data(self) -> Data:
        """Pass data to the sub-rewarders.

        :meta private:
        """
        for rewarder, data in zip(self.rewarders, self.data.data):
            rewarder.data = data

    def __init__(self, *rewarders: GlobalReward) -> None:
        """Rewarder for composed data types.

        This type can be automatically constructed by passing a tuple of rewarders to
        the environment constructor's `reward` argument.

        Args:
            rewarders: Global rewarders to compose.
        """
        super().__init__()
        self.rewarders = rewarders

    def __getattr__(self, name: str):
        """Search for an attribute in the rewarders."""
        for rewarder in self.rewarders:
            if hasattr(rewarder, name):
                return getattr(rewarder, name)
        raise AttributeError(
            f"No GlobalReward in ComposedReward has attribute '{name}'"
        )

    def reset_pre_sim_init(self) -> None:
        """Handle resetting for all rewarders."""
        super().reset_pre_sim_init()
        for rewarder in self.rewarders:
            rewarder.reset_pre_sim_init()

    def reset_post_sim_init(self) -> None:
        """Handle resetting for all rewarders."""
        super().reset_post_sim_init()
        for rewarder in self.rewarders:
            rewarder.reset_post_sim_init()

    def reset_overwrite_previous(self) -> None:
        """Handle resetting for all rewarders."""
        super().reset_overwrite_previous()
        for rewarder in self.rewarders:
            rewarder.reset_overwrite_previous()

    def link_scenario(self, scenario: Scenario) -> None:
        """Link the rewarder to the scenario."""
        super().link_scenario(scenario)
        for rewarder in self.rewarders:
            rewarder.link_scenario(scenario)

    def initial_data(self, satellite: Satellite) -> ComposedData:
        """Furnish the DataStore with :class:`ComposedData`."""
        return ComposedData(
            *[rewarder.initial_data(satellite) for rewarder in self.rewarders]
        )

    def create_data_store(self, satellite: Satellite) -> None:
        """Create a :class:`CompositeDataStore` for a satellite."""
        satellite.data_store = ComposedDataStore(
            satellite,
            *[r.data_store_type for r in self.rewarders],
            initial_data=self.initial_data(satellite),
            data_store_kwargs=[r.data_store_kwargs for r in self.rewarders],
        )
        self.cum_reward[satellite.name] = 0.0
        for rewarder in self.rewarders:
            rewarder.cum_reward[satellite.name] = 0.0

    def calculate_reward(
        self, new_data_dict: dict[str, ComposedData]
    ) -> dict[str, float]:
        """Calculate reward for each data type and combine them."""
        data_len = len(list(new_data_dict.values())[0].data)

        for data in new_data_dict.values():
            assert len(data.data) == data_len

        reward = {}
        if data_len != 0:
            for i, rewarder in enumerate(self.rewarders):
                reward_i = rewarder.calculate_reward(
                    {sat_id: data.data[i] for sat_id, data in new_data_dict.items()}
                )

                # Logging
                nonzero_reward = {k: v for k, v in reward_i.items() if v != 0}
                if len(nonzero_reward) > 0:
                    logger.info(f"{type(rewarder).__name__} reward: {nonzero_reward}")

                for sat_id, sat_reward in reward_i.items():
                    reward[sat_id] = reward.get(sat_id, 0.0) + sat_reward
                    rewarder.cum_reward[sat_id] += sat_reward
        return reward

    def reward(self, new_data_dict: dict[str, ComposedData]) -> dict[str, float]:
        """Return combined reward calculation and update data."""
        reward = super().reward(new_data_dict)
        self.pass_data()
        return reward

    def is_truncated(self, satellite: Satellite) -> bool:
        """Check if the episode is truncated by any rewarder."""
        return any(rewarder.is_truncated(satellite) for rewarder in self.rewarders)

    def is_terminated(self, satellite) -> bool:
        """Check if the episode is terminated by any rewarder."""
        return any(rewarder.is_terminated(satellite) for rewarder in self.rewarders)


__doc_title__ = "Data Composition"
__all__ = ["ComposedReward", "ComposedDataStore", "ComposedData"]
